#include <windows.h>
#include <stdio.h>
#include <winsock.h>
#include <time.h>
#include <string.h>
#include <commctrl.h>
#include <sys/timeb.h>
#include "resource.h"
#include "hash.h"
#include "main.h"
#include "events.h"
#include "msg.h"
#include "chat.h"

#define CHAT_PROTOCOL	2
#define CHAT_TIMEOUT	45
#define PING_PERIOD		10

#define BUF_SIZE		1024


#define SVID HashSelectInt(data,"Chatclient","svid")
#define CLID HashSelectInt(data, "Chatclient","clid")

void CHAT_Login(void);

int WINAPI CHAT_UsersProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam);



int parse_input (char* buf, void* owner, char* name)
{
  int i, j, offset;
  char key[BUF_SIZE], value[BUF_SIZE];

  while (*buf != '\\' && *buf)
    buf++;

  i = 0; offset = 0;
  while (*buf) {
    buf++; // skip the backslash
    j = 0;
    while (*buf && *buf != '\\') {
      if (!offset)
		  key[j] = *buf;
	  else value[j] = *buf;
      j++; buf++;
    }
    if (!offset)
		key[j] = 0;
	else value[j] = 0;
    if (offset) {
		HashInsertText(owner,name,key,value);
      offset = 0; i++;
    }
    else offset = 1;
  }
  return i;
}



void CHAT_Cleanup(CHAT_DATA* data) {
	SOCKADDR_IN daddr;
	char addr[128];
	char buf[256];

	EVENTS_Cancel(data->event);
		DestroyWindow(data->hwnd);

	
	stack_empty(data);

	if (data->socket)
	{
		daddr.sin_family = AF_INET;
		sscanf(data->server,"%s",addr);
		daddr.sin_addr.s_addr = inet_addr(addr);
		daddr.sin_port = htons(25555);
		sprintf(buf,"\xff\xff\xff\xff" "disconnect \\genid\\%d",data->genid);
		sendto(data->socket,buf,strlen(buf),0,
			(SOCKADDR*)&daddr,sizeof(daddr));
		closesocket(data->socket);
	}

	HashInsertInt(HASH_GLOBAL,"Status","ChatActive",0);

	if (data->users && HSIG("Settings","SideChat") == 0 )
		DestroyWindow(data->users);
	else SendMessage(data->users,WM_SIZE,0,0);
	
	free(data);


	return;
}

int WINAPI CHAT_LoginWaitProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
	if (msg == WM_COMMAND && wparam == IDB_ABORT) {
		CHAT_Cleanup((CHAT_DATA*)GetWindowLong(hwnd,GWL_USERDATA));
	}
	return 0;
}



#define SET_LASTRESP (data->last_response = time(NULL))

#define CLEAR_MSG (data->next_msg[0] = 0)

userdata_t* FindUserWithID(CHAT_DATA* data, int id)
{
	userdata_t* ud;
	
	for (ud=HashReset(data,"UserDataPtrs",HASH_PTR);;
			ud =HashNext(data,"UserDataPtrs",HASH_PTR))
		if (ud->userid == id )
			return ud;

	return NULL;
}
			
void CHAT_Daemon(CHAT_DATA* data) {
	char buf[BUF_SIZE];
	char msg[BUF_SIZE];
	char addr[128];
	SOCKADDR_IN daddr;
	FD_SET fs;
	userdata_t* ud;
	struct timeval tv;
	int i, index;
	char* ptr;
	RECT r;
	struct _timeb tb;

	if (!data->socket) {
		//fixme - error check
		data->socket = socket(AF_INET,SOCK_DGRAM,IPPROTO_UDP);
		data->hwnd = CreateDialog(PROG_HINSTANCE,MAKEINTRESOURCE(IDD_GWAIT),PROG_HWND,CHAT_LoginWaitProc);
		SetWindowLong(data->hwnd,GWL_USERDATA,(LPARAM)data);
		SetDlgItemText(data->hwnd,IDT_WAIT,"Server Query is attempting to connect to the selected chat server");
		ShowWindow(data->hwnd,SW_SHOW);
		data->stage = 0;
		SET_LASTRESP;
	}

	FD_ZERO(&fs);
	FD_SET(data->socket,&fs);

	tv.tv_sec = 0; tv.tv_usec = 0;
	while ( select(0,&fs,NULL,NULL,&tv) > 0)
	{
		tv.tv_sec = 0; tv.tv_usec = 0;

		i = sizeof(SOCKADDR_IN);
		i = recvfrom(data->socket,buf,BUF_SIZE-10,0,(SOCKADDR*)&daddr,&i);

		if (i>0 && memcmp(buf,"\xff\xff\xff\xff",4) == 0)
		{
			buf[i] = 0;
			ptr = buf + 4;
			sscanf(ptr,"%s",msg); // get the msg type
	
			HashDeleteAll(data,"ChatClient");
			parse_input(ptr,data,"ChatClient");

			if (sq_stricmp(msg,"pingResponse") == 0) {
				stack_delmessage(data,stack_search(data,"ping ") );
				SET_LASTRESP;

				if (data->stage == 0) // login
				{
					data->genid = rand() % 99999;
					sprintf(buf,
						"login \\username\\%s\\password\\%s\\protocol\\%d",
						data->username,
						data->password,
						CHAT_PROTOCOL);
					stack_add(data,buf,0,0);
					data->stage ++;
				}
			}

			if (data->stage == 1 && sq_stricmp(msg,"loginResponse") == 0)
			{
				stack_delmessage(data,stack_search(data,"login") );
				SET_LASTRESP;
				
				if (strcmp(HashSelectText(data,"ChatClient","valid"),"no") == 0) {
					CHAT_Cleanup(data);
					MSG_DialogBox(NULL,NULL,"Invalid login information or access denied.\n\n"
							"The server you have attempted to connect to has rejected your login "
							"information.  Please confirm that your username and password are correct "
							"and that you have specified the correct server.",IDD_ERROR);
					CHAT_Login();
					return;
				}
				else if (strcmp(HashSelectText(data,"ChatClient","valid"),"yes") == 0)
				{
					data->stage ++; // fully logged in
					data->id = HashSelectInt(data,"ChatClient","ID");
					data->cl_msg = 1;

					ShowWindow(data->hwnd,SW_HIDE); // hide the 'busy' window
					HashInsertPtr(HASH_GLOBAL,"Status","ChatActive",data);
					if (HSIG("Settings","SideChat") ==0)
						data->users = CreateDialog(PROG_HINSTANCE,MAKEINTRESOURCE(IDD_CHAT),PROG_HWND,CHAT_UsersProc);
					else {
						data->users = PROG_HWND;
						SendMessage(PROG_HWND,WM_SIZE,0,0);
					}
					TreeView_SetImageList(GetDlgItem(data->users,IDC_USERS), HashSelectPtr(NULL,"Global","ImageList"), TVSIL_NORMAL); 

					SetWindowLong(data->users,GWL_USERDATA,(LPARAM)data);
					
					if (HSIG("Settings","SideChat") == 0)
					{
						GetWindowRect(PROG_HWND,&r);
						ShowWindow(data->users,SW_SHOWNORMAL);
						SetWindowPos(data->users,NULL,r.right-50,r.top+100,0,0,SWP_NOSIZE|SWP_NOZORDER);
					}

					//CLEAR_MSG;
					return;
				}
			}

			if (sq_stricmp(msg,"error") == 0) {
				stack_delmessage(data,stack_clid(data,CLID) );
				SET_LASTRESP;
				MSG_DialogBox(NULL,NULL,HashSelectText(data,"chatClient","message"),IDD_ERROR);
			}

			// Packets that are for our information only (SVID is only 
			// enabled on comms beginning at the server)
			// be acted upon once
			//
			// stack_add updates data->svid for us
			if (SVID > data->svid)
			{
				if (sq_stricmp(msg,"recvMessageUpdate") == 0)
				{
					SET_LASTRESP;
					stack_add(data,"recvMessage ",SVID,1);
				}
				if (sq_stricmp(msg,"userlistUpdate") == 0)
				{
					SET_LASTRESP;
					stack_add(data,"userlist ",SVID,1);
				}
			}

			// Response to a previous request (this is a reply to a comms -
			// the msg should still be in the stack) if it hasn't been
			// acted upon
			//
			if ((index = stack_clid(data,CLID)) != -1)
			{
				if (sq_stricmp(msg,"sendMessageResponse") == 0)
				{
					stack_delmessage(data,index);
					SET_LASTRESP;
				}
				if (sq_stricmp(msg,"recvMessageResponse") == 0)
				{
					stack_delmessage(data,index);
					SET_LASTRESP;
					//CHAT_RecvMessage(data);
					if (GetActiveWindow() != GetForegroundWindow())
						FlashWindow(PROG_HWND,0);
					ud = FindUserWithID(data,HashSelectInt(data,"ChatClient","from"));
					if (ud)
						incstack_addmessage(&ud->incstack,ptr);

					PlaySound("sounds/incoming.wav",NULL,SND_ASYNC|SND_FILENAME|SND_NOWAIT);
				}
				if (sq_stricmp(msg,"userlistResponse") == 0)
				{
					stack_delmessage(data,index);
					SET_LASTRESP;
					CHAT_UpdateList(data);
				}
			}
		}

	}

	if (data->last_response + CHAT_TIMEOUT < time(NULL))
	{
		CHAT_Cleanup(data);
		MSG_DialogBox(NULL,NULL,"The connection to the selected chat server has timed out.",IDD_ERROR);
		return;
	}

	_ftime(&tb);
	if (data->users)
	{
		if (data->last_icons == 0 && tb.millitm < 500)
		{
			data->last_icons = 500;
			CHAT_UpdateIcons(data);
		}
		else if (data->last_icons == 500 & tb.millitm >= 500)
		{
			data->last_icons = 0;
			CHAT_UpdateIcons(data);
		}
	}
		
	
	if (time(NULL) % PING_PERIOD == 0 && data->last_ping < time(NULL))
	{
		stack_add(data,"ping ",0,0);
		data->last_ping = time(NULL);
	}

	if (data->stack_size) {
		if (data->stack[0]->msg[0] != '\xff')
			strcpy(buf,"\xff\xff\xff\xff");
		else strcpy(buf,"");
		sprintf(buf+strlen(buf),"%s\\genid\\%d",data->stack[0]->msg,
			data->genid);
		if (data->stack[0]->svid)
			sprintf(buf+strlen(buf),"\\svid\\%d",data->stack[0]->svid);
		if (data->stack[0]->clid)
			sprintf(buf+strlen(buf),"\\clid\\%d",data->stack[0]->clid);

		daddr.sin_family = AF_INET;
		sscanf(data->server,"%s",addr);
		daddr.sin_addr.s_addr = inet_addr(addr);
		daddr.sin_port = htons(25555);
		if (sendto(data->socket,buf,strlen(buf),0,(SOCKADDR*)&daddr,sizeof(daddr))
			< 0)
			CHAT_Cleanup(data);
	}
	return;
}

int WINAPI CHAT_LoginProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
	HWND tmp;
	char buf[128];
	CHAT_DATA* data;

	if (msg == WM_INITDIALOG)
	{
		tmp = GetDlgItem(hwnd,IDC_SERVER);
		SendMessage(tmp,CB_ADDSTRING,0,(LPARAM)"202.139.235.147 - QGL Server");
		SendMessage(tmp,CB_SETCURSEL,0,0);
		if (HashSelectInt(HASH_GLOBAL,"Settings","ChatSaveLogin"))
		{
			SetDlgItemText(hwnd,IDT_USERNAME,HashSelectText(HASH_GLOBAL,"Settings","ChatUsername") );
			SetDlgItemText(hwnd,IDT_PASSWORD,HashSelectText(HASH_GLOBAL,"Settings","ChatPassword") );
			SendMessage(GetDlgItem(hwnd, IDC_USERPASS), BM_SETCHECK, BST_CHECKED, 0);
		}
		if (HashSelectInt(HASH_GLOBAL,"Settings","ChatStart"))
			SendMessage(GetDlgItem(hwnd,IDC_START),BM_SETCHECK,BST_CHECKED,0);
		else EnableWindow(GetDlgItem(hwnd,IDC_AUTOLOGIN),0);

		if (HashSelectInt(HASH_GLOBAL,"Settings","ChatAutologin"))
			SendMessage(GetDlgItem(hwnd,IDC_AUTOLOGIN),BM_SETCHECK,BST_CHECKED,0);
	}
	if (msg == WM_COMMAND)
	{
		if (LOWORD(wparam) == IDC_START)
		{
			if (IsDlgButtonChecked(hwnd, IDC_START) == BST_CHECKED)
				EnableWindow(GetDlgItem(hwnd,IDC_AUTOLOGIN),1);
			else EnableWindow(GetDlgItem(hwnd,IDC_AUTOLOGIN),0);
			return 0;
		}

		if (wparam == IDOK) {
			data = calloc(1,sizeof(CHAT_DATA));

			GetDlgItemText(hwnd,IDC_SERVER,buf,128);
			strcpy(data->server,buf);
			GetDlgItemText(hwnd,IDT_USERNAME,buf,128);
			
			strcpy(data->username,buf);
			GetDlgItemText(hwnd,IDT_PASSWORD,buf,128);
			
			strcpy(data->password,buf);

			if (SendMessage(GetDlgItem(hwnd, IDC_USERPASS), BM_GETCHECK, 0, 0) 
				== BST_CHECKED)
			{
				HashInsertInt(HASH_GLOBAL,"Settings","ChatSaveLogin",1);
				HashInsertText(HASH_GLOBAL,"Settings","ChatServer",data->server);
				HashInsertText(HASH_GLOBAL,"Settings","ChatUsername",data->username);
				HashInsertText(HASH_GLOBAL,"Settings","ChatPassword",data->password);
			}
			else
			{
				HashInsertInt(HASH_GLOBAL,"Settings","ChatSaveLogin",0);
				HashDelete(HASH_GLOBAL,"Settings","ChatServer");
				HashDelete(HASH_GLOBAL,"Settings","ChatUsername");
				HashDelete(HASH_GLOBAL,"Settings","ChatPassword");
			}

			if (SendMessage(GetDlgItem(hwnd, IDC_START), BM_GETCHECK, 0, 0) 
					== BST_CHECKED)
				HashInsertInt(HASH_GLOBAL,"Settings","ChatStart",1);
			else HashInsertInt(HASH_GLOBAL,"Settings","ChatStart",0);

			if (SendMessage(GetDlgItem(hwnd, IDC_AUTOLOGIN), BM_GETCHECK, 0, 0) 
					== BST_CHECKED)
				HashInsertInt(HASH_GLOBAL,"Settings","ChatAutologin",1);
			else HashInsertInt(HASH_GLOBAL,"Settings","ChatAutologin", 0);

			EndDialog(hwnd,(int)data);
		}
		if (wparam == IDCANCEL)
			EndDialog(hwnd,0);
	}
	if (msg == WM_CLOSE)
		EndDialog(hwnd,0);

	return 0;
}
void CHAT_Login(void) {
	CHAT_DATA* data;

	if ((data = HSPG("Status","ChatActive")) != NULL) {
		//MSG_DialogBox(PROG_HINSTANCE,PROG_HWND,"You are already connected to a chat server.",IDD_ERROR);
		CHAT_Cleanup(data);
		return;
	}

	data = NULL;
	if (HashSelectInt(HASH_GLOBAL,"Settings","ChatAutologin") &&
		HashSelectInt(HASH_GLOBAL,"Status","ChatStartup"))
	{
		HashDelete(HASH_GLOBAL,"Status","ChatStartup");
		data = calloc(1,sizeof(CHAT_DATA));
		strcpy(data->server,HashSelectText(HASH_GLOBAL,"Settings","ChatServer"));
		strcpy(data->username,HashSelectText(HASH_GLOBAL,"Settings","ChatUsername"));
		strcpy(data->password,HashSelectText(HASH_GLOBAL,"Settings","ChatPassword"));
	}

	if (data || (data = (CHAT_DATA*)DialogBox(PROG_HINSTANCE,MAKEINTRESOURCE(IDD_CHATLOGIN),
		PROG_HWND,CHAT_LoginProc)) != NULL)
	{
		stack_add(data,"ping ",0,0);
		data->event = EVENTS_Register(CHAT_Daemon,data,0,(float)0.25);
	}
	return;
}