#include <windows.h>
#include <commctrl.h>
#include "chat.h"
#include "resource.h"
#include "hash.h"
#include <stdio.h>
#include "msg.h"
#include "main.h"

#define MSG_SIZE 800
#define BUF_SIZE 1024


void CHAT_UpdateIcons(CHAT_DATA* data)
{
	userdata_t* ud;
	TV_ITEM tvi;
	int r;
	static int toggle;

	toggle = (toggle+1)%2;

	for (ud = HashReset(data,"UserDataPtrs",HASH_PTR);ud;
		ud = HashNext(data,"UserDataPtrs",HASH_PTR))
	{
		tvi.hItem = ud->hTreeItem;
		tvi.iImage = tvi.iSelectedImage = HashSelectInt(NULL,"Indexes","Blank");
		if (incstack_size(&ud->incstack))
		{
			if (toggle)
		      tvi.iImage = tvi.iSelectedImage = HashSelectInt(NULL,"Indexes","Message");
			else tvi.iImage = tvi.iSelectedImage = HashSelectInt(NULL,"Indexes","None");
		}
	
		tvi.mask = TVIF_IMAGE | TVIF_SELECTEDIMAGE;
		r= TreeView_SetItem(GetDlgItem(data->users,IDC_USERS),&tvi);
	}

	return;
}

		

void CHAT_UpdateList(CHAT_DATA* data)
{
	TV_INSERTSTRUCT tv;
	userdata_t *ud;
	HTREEITEM root;

	HWND hwnd = GetDlgItem(data->users,IDC_USERS);
	int i, total;
	char buf[128];

	for (ud = HashReset(data,"UserDataPtrs",HASH_PTR);ud;
			ud = HashNext(data,"UserDataPtrs",HASH_PTR))
		ud->inuse = 0;
		
	TreeView_DeleteAllItems(hwnd);
	
	// Create root item
	tv.hParent = TVI_ROOT;
	tv.hInsertAfter = TVI_LAST;
	tv.item.pszText = "Online Users";
	tv.item.iImage = HashSelectInt(NULL,"Indexes","Open");
	tv.item.iSelectedImage = HashSelectInt(NULL,"Indexes","Open");
	tv.item.lParam = (long)NULL;

	tv.item.mask = TVIF_TEXT | TVIF_PARAM | TVIF_IMAGE | TVIF_SELECTEDIMAGE;
	root = TreeView_InsertItem(hwnd,&tv);
	HIP(data,"status","rootitem",root);

	total = HashSelectInt(data,"ChatClient","total");

	for (i=0;i<total;i++)
	{
		sprintf(buf,"user%d",i);
		for (ud = HashReset(data,"UserDataPtrs",HASH_PTR);ud;
				ud = HashNext(data,"UserDataPtrs",HASH_PTR))
			if (ud->userid == HashSelectInt(data,"ChatClient",buf))
				break;

		if (!ud) {
			ud = calloc(sizeof(userdata_t),1);
			HashInsertPtr(data,"UserDataPtrs",HASH_ANY, ud);
		}
		ud->inuse = 1;
		
		sprintf(buf,"user%d",i);
		ud->userid = HashSelectInt(data,"ChatClient",buf);
		sprintf(buf,"name%d",i);
		strcpy(ud->name, HashSelectText(data,"ChatClient",buf) );
		sprintf(buf,"clan%d",i);
		strcpy(ud->clan, HashSelectText(data,"ChatClient",buf) );
		

		tv.hParent = root;
		tv.hInsertAfter = TVI_LAST;
		tv.item.pszText = ud->name;
		tv.item.lParam = (LPARAM)ud;
		tv.item.iImage = HashSelectInt(NULL,"Indexes","Blank");
		tv.item.iSelectedImage = HashSelectInt(NULL,"Indexes","Blank");

	/*if (GetServerType(svr->type)->iImage) 
		tv.item.iImage = tv.item.iSelectedImage = GetServerType(svr->type)->iImage; */

		tv.item.mask = TVIF_TEXT | TVIF_PARAM | TVIF_IMAGE | TVIF_SELECTEDIMAGE;
		ud->hTreeItem = TreeView_InsertItem(hwnd,&tv);

		/*tv.item.hItem = svr->hTreeItem;
		TreeView_SetItem(hwnd,&(tv.item));*/
	}
	TreeView_Expand(hwnd,root,TVE_EXPAND);

	for (ud = HashReset(data,"UserDataPtrs",HASH_PTR);ud;
			ud = HashNext(data,"UserDataPtrs",HASH_PTR))
		if (ud->inuse == 0)
		{
			HashDeleteValue(data,"UserDataPtrs",ud,HASH_PTR);
			free(ud);
			ud = HashReset(data,"UserDataPtrs",HASH_PTR);
		}

	return;
}

typedef struct 
{
	CHAT_DATA* data;
	userdata_t* user;
	HWND hwnd;
} message_t;


char* CHAT_Decode(char* str)
{
	static char decoded[BUF_SIZE];
	int i, j, k;
	char temp[3];

	for (i=0,j=0;i<(int)strlen(str);j++)
	{
		if (str[i] == '%')
		{
			temp[0] = str[i+1];
			temp[1] = str[i+2];
			temp[2] = 0;
			sscanf(temp,"%x",&k);
			decoded[j] = k;
			i+=3;
		}
		else 
		{
			decoded[j] = str[i];
			i++;
		}
	}

	decoded[j] = 0;
	return decoded;
}

char* CHAT_Encode(char* str)
{
	static char encoded[BUF_SIZE];
	char *letters = "%\\\n\t\r";
	int i, j, k;
	int encode;

	for (i=0,j=0;i<(int)strlen(str) && j < MSG_SIZE-4;i++)
	{
		encode = 0;
		for (k=0;k<(int)strlen(letters);k++)
			if (str[i] == letters[k])
				encode = 1;
		if (str[i] < 32)
			encode = 1;

		if (encode)
		{
			sprintf(encoded+j,"%%%02x",(int)str[i]);
			j+= 3;
		}
		else 
		{
			encoded[j] = str[i];
			j++;
		}
	}

	encoded[j] = 0;
	return encoded;
}


int WINAPI CHAT_SendMessageProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
	message_t* message;
	char buf[1024];
	char strMesg[1024];

	if (msg == WM_INITDIALOG)
		return -1;

	if (msg == WM_COMMAND && wparam == IDOK)
	{
		message = (message_t*)GetWindowLong(hwnd,GWL_USERDATA);
		sprintf(buf,"sendMessage \\to\\%d\\message\\",message->user->userid);
		GetDlgItemText(hwnd,IDT_MESSAGE,strMesg,950);
		strcat(buf,CHAT_Encode(strMesg));
		stack_add(message->data,buf,0,1);
		msg = WM_CLOSE;
	}
	if ((msg == WM_COMMAND && wparam == IDCANCEL) || msg == WM_CLOSE) {
		message = (message_t*)GetWindowLong(hwnd,GWL_USERDATA);
		free(message);
		DestroyWindow(hwnd);
	}
	return 0;
}



void CHAT_SendMessage(CHAT_DATA* data, void* to)
{
	HWND hwnd;
	message_t* message;
	char buf[128];

	hwnd = CreateDialog(PROG_HINSTANCE,MAKEINTRESOURCE(IDD_CHATMSG),
		PROG_HWND,CHAT_SendMessageProc);

	sprintf(buf,"Send to: %s", ((userdata_t*)to)->name);
	SetDlgItemText(hwnd,IDT_SENDTO,buf);

	message = calloc(1,sizeof(message_t));
	message->hwnd = hwnd;
	message->data = data;
	message->user = to;
	SetWindowLong(hwnd,GWL_USERDATA,(long)message);

	ShowWindow(hwnd,SW_SHOWNORMAL);
	HashInsertPtr(HASH_GLOBAL,"Status","NewWindow",hwnd);

	return;
}

int WINAPI CHAT_RecvMessageProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
	message_t* message;
	//char buf[1024];
	HDC hdc;
	//COLORREF res;

	if (msg == WM_INITDIALOG)
	{
		//EnableWindow(GetDlgItem(hwnd,IDT_MESSAGE),0);
		//SetBkColor( GetDC(GetDlgItem(hwnd,IDT_MESSAGE)), RGB(255,0,0) );
		SendMessage(GetDlgItem(hwnd,IDT_MESSAGE),EM_SETREADONLY,1,0);
		hdc = GetDC(GetDlgItem(hwnd,IDT_MESSAGE) );
		SetTextColor(hdc, RGB(0,255,0) );
		SetBkColor(hdc,RGB(255,0,0) );
		return 0;
	}

	if (msg == WM_CTLCOLOREDIT && lparam == (long)GetDlgItem(hwnd,IDT_MESSAGE))
	{
		//	printf("Hurrah");
		hdc = (void*)wparam;
		//SelectObject(hdc, CreateSolidBrush(RGB(255,0,0) ));
		SetTextColor(hdc, RGB(0,255,0) );
		SetBkColor(hdc,RGB(255,0,0) );
		return (int)CreateSolidBrush(RGB(255,0,0) );
	}

	/*if (msg == WM_COMMAND && wparam == IDOK)
	{
		message = (message_t*)GetWindowLong(hwnd,GWL_USERDATA);
		sprintf(buf,"sendMessage \\to\\%d\\message\\",message->user->userid);
		GetDlgItemText(hwnd,IDT_MESSAGE,buf+strlen(buf),950);
		stack_add(message->data,buf,0,1);
		msg = WM_CLOSE;
	}*/
	if (msg == WM_COMMAND && wparam == IDOK)
	{
		message = (message_t*)GetWindowLong(hwnd,GWL_USERDATA);
		DestroyWindow(hwnd);
		CHAT_SendMessage(message->data,message->user);
		free(message);
	}
	if ((msg == WM_COMMAND && wparam == IDCANCEL) || msg == WM_CLOSE) {
		message = (message_t*)GetWindowLong(hwnd,GWL_USERDATA);
		free(message);
		DestroyWindow(hwnd);
	}
	return 0;
}

WNDPROC oldEditProc;
int WINAPI HookedEditProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{

	if (msg == WM_SETFOCUS)
		return 0;

	return CallWindowProc(oldEditProc,hwnd,msg,wparam,lparam);
	
}


void CHAT_RecvMessage(CHAT_DATA* data)
{
	HWND hwnd;
	char buf[128];
	message_t* message;
	userdata_t* ud;

	hwnd = CreateDialog(PROG_HINSTANCE,MAKEINTRESOURCE(IDD_CHATMSG),
		PROG_HWND,CHAT_RecvMessageProc);

	for (ud=HashReset(data,"UserDataPtrs",HASH_PTR);;
			ud =HashNext(data,"UserDataPtrs",HASH_PTR))
		if (ud->userid == HashSelectInt(data,"ChatClient","from") )
			break;
	
	if (ud)
		strcpy(buf,ud->name);
	else strcpy(buf,"Anonymous User");
	sprintf(buf+strlen(buf)," writes:");
	
	SetDlgItemText(hwnd,IDT_SENDTO,buf);
	SetDlgItemText(hwnd,IDCANCEL,"Close");
	SetDlgItemText(hwnd,IDOK,"Reply");
	
	SetWindowText(hwnd,"Incoming Message");						

	/*oldEditProc = SetWindowLong(GetDlgItem(hwnd,IDT_MESSAGE), 
		GWL_WNDPROC, (long)HookedEditProc);*/

	ShowWindow(hwnd,SW_SHOWNORMAL);

	//EnableWindow(GetDlgItem(hwnd,IDT_MESSAGE),0);
SetDlgItemText(hwnd,IDT_MESSAGE,CHAT_Decode(HashSelectText(data,"ChatClient","message")));
	
	message = calloc(1,sizeof(message_t));

	message->data = data;
	message->user = ud;
	SetWindowLong(hwnd,GWL_USERDATA,(long)message);

	HashInsertPtr(HASH_GLOBAL,"Status","NewWindow",hwnd);
	SetFocus(GetDlgItem(hwnd,IDOK));

	
	
	return;
}

HTREEITEM FUNC_GetHitItem (HWND hwnd) {
	static int x, y, temp;
	TV_HITTESTINFO info;
	RECT rect;
	
	temp = GetMessagePos();
	x = LOWORD(temp);
	y = HIWORD(temp);

	GetWindowRect(hwnd,&rect);
	info.pt.x = x - rect.left;
	info.pt.y = y - rect.top;

	return TreeView_HitTest(hwnd,&info);
}

void CHAT_ProcessStack(CHAT_DATA* data, userdata_t* ud)
{
	char msg[128];
	char* ptr = incstack_getmessage(&ud->incstack,0);
	
	sscanf(ptr,"%s",msg);
	HashDeleteAll(data,"ChatClient");
	parse_input(ptr,data,"ChatClient");

	incstack_delmessage(&ud->incstack,0);

	if (sq_stricmp(msg,"recvMessageResponse") == 0)
		CHAT_RecvMessage(data);

	return;
}

int WINAPI CHAT_UsersProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
	NMHDR* hdr;
	//NM_TREEVIEW* tv;
	TV_ITEM tvi;
	CHAT_DATA* data;
	userdata_t* ud;

	data = (CHAT_DATA*)GetWindowLong(hwnd,GWL_USERDATA);

	if ((msg == WM_COMMAND && wparam == IDB_ABORT) || msg == WM_CLOSE) {
		CHAT_Cleanup(data);
	}

	if (msg == WM_NOTIFY)
	{
		hdr = (NMHDR*)lparam;
		/*if (hdr->idFrom == IDC_USERS && hdr->code == TVN_ITEMEXPANDED)
		{
			tv = (NM_TREEVIEW*)lParam;
			tv->itemNew.mask = TVIF_IMAGE | TVIF_SELECTEDIMAGE;
			if (tv->action == 2) {
				tv->itemNew.iImage = HashSelectInt(NULL,"Indexes","Open");
				tv->itemNew.iSelectedImage = HashSelectInt(NULL,"Indexes","Open");
			}
			else {
				tv->itemNew.iImage = HashSelectInt(NULL,"Indexes","Closed");
				tv->itemNew.iSelectedImage = HashSelectInt(NULL,"Indexes","Closed");
			}
			TreeView_SetItem(HashSelect(HASH_GLOBAL,"Global","ServersTree",HASH_PTR),&tv->itemNew);
		}
		if (hdr->idFrom == IDC_TREE1 && hdr->code == NM_RCLICK)
		{
			DisplaySCM();
		}*/

		if (hdr->idFrom == IDC_USERS && hdr->code == NM_DBLCLK)
		{
			tvi.hItem = FUNC_GetHitItem(GetDlgItem(hwnd,hdr->idFrom));
			tvi.mask = TVIF_PARAM;
			TreeView_GetItem(GetDlgItem(hwnd,hdr->idFrom),&tvi);
			if (!tvi.lParam)
				return 0;
			ud = (userdata_t*)tvi.lParam;
			if (incstack_size(&ud->incstack))
				CHAT_ProcessStack(data, ud);
			else CHAT_SendMessage(data, (void*)tvi.lParam);
			return 1;
		}
	}
	return 0;
}
