/*
******************************************************************
Server Query; a free games server monitoring tool
Copyright (C) 1998-2002  Nathan O'Sullivan

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
******************************************************************
*/
#include <windows.h>
#include <stdio.h>
#include "resource.h"

#include "servers.h"
#include "main.h"
#include "msg.h"
#include "systray.h"
#include "servinfo.h"
#include "list.h"
#include "scheme.h"

list_t* GetLANList(void);


char* current_path;
#define NUM_REFRESH 7
char* refresh_strs[] = { "1sec","2sec","5sec","10sec","15sec", "20sec", "30sec" };
int refresh_values[] = {1, 2, 5, 10, 15,20, 30 };

#define NUM_TIMEOUT 8
char* timeout_strs[] = { "2sec","5sec","10sec","15sec", "20sec", "30sec", "45sec", "60sec" };
int timeout_values[] = {2, 5, 10, 15,20,30,45,60 };

#define NUM_REFSELLIST 10
char* refsellist_strs[] = { "Never","5sec","10sec","15sec", "30sec", "60sec", "2min", "3min","5min","10min" };
int refsellist_values[] = {-1,5,10,15,30,60,120,180,300,600 };

typedef struct {
	int id;
	char* name;
	int reverse;
	int rebuildWindow;
} BOOL_SETTINGS;

BOOL_SETTINGS boolSettings[] = {
	{ IDC_EXIT, "ExitOnConnect", 0, 0 },
	/* { IDC_SHOWGAMETYPE, "ShowGameTypeOnLan", 0, 0 }, */
	{ IDC_PINGALL, "PingAllNetServers", 0, 0 },
	/* { IDC_LANONLY, "NoLAN", 0}, */
	{ IDC_REFRESHPUPSERVERS,"RefreshPupServers", 0, 0 },
	{ IDC_FOREGROUNDONLY, "PingWhileBackground", 1, 0},
	{ IDC_SORTLAN, "LANSortbyGame", 0, 0 },
	{ IDC_SYSTRAY, "Systray", 0, 0},
	{ IDC_MOUSEWHEEL, "MouseWheel", 0, 0},
	{ IDC_DUALPANE, "DualPane", 1, 1},
	{ IDC_SIDECHAT, "SideChat", 0, 1},
	{ IDC_UPDATENAMES, "UpdateNames", 0, 0 },
	{ IDC_EXTRAINFO, "ExtraServInfo", 0, 2 },
	{ IDC_LISTALLSERVERS, "ListAllServers", 0, 1 },
	{ IDC_SHOWLANLIST,	"ShowLANList", 0, 3},
	{ IDC_SCHEMES, "EnableSchemes", 0, 4 },
	{ 0, NULL, 0, 0}
};

int GENCFG_Close(HWND hwnd) 
{
	char buf[128];
	int i, refresh, timeout, lan_refresh, lan_timeout;
	int old, value, rebuild = 0;
	
	SendMessage(GetDlgItem(hwnd, IDC_REFRESH), WM_GETTEXT, 128, (LPARAM)buf);
	for (i=0;i<NUM_REFRESH;i++)
		if (!strcmp(refresh_strs[i], buf))
			refresh = refresh_values[i];

		SendMessage(GetDlgItem(hwnd, IDC_TIMEOUT), WM_GETTEXT, 128, (LPARAM)buf);
	for (i=0;i<NUM_TIMEOUT;i++)
		if (!strcmp(timeout_strs[i], buf))
			timeout = timeout_values[i];

	SendMessage(GetDlgItem(hwnd, IDC_LANREFRESH), WM_GETTEXT, 128, (LPARAM)buf);
	for (i=0;i<NUM_REFRESH;i++)
		if (!strcmp(refresh_strs[i], buf))
			lan_refresh = refresh_values[i];

	SendMessage(GetDlgItem(hwnd, IDC_LANTIMEOUT), WM_GETTEXT, 128, (LPARAM)buf);
	for (i=0;i<NUM_TIMEOUT;i++)
		if (!strcmp(timeout_strs[i], buf))
			lan_timeout = timeout_values[i];



	if (refresh >= timeout || lan_refresh > lan_timeout)
	{
		MSG_DialogBox(NULL, NULL, "Sorry, the refresh rate must be less than the timeout length.", IDD_ERROR);
		return 0;
	}
	HashInsert(HASH_GLOBAL,"Settings","Refresh",(HASH_PARAM)refresh,HASH_INT);
	HashInsert(HASH_GLOBAL,"Settings","Timeout",(HASH_PARAM)timeout,HASH_INT);
	HashInsert(HASH_GLOBAL,"Settings","LANRefresh",(HASH_PARAM)lan_refresh,HASH_INT);
	HashInsert(HASH_GLOBAL,"Settings","LANTimeout",(HASH_PARAM)lan_timeout,HASH_INT);
	HIIG("Settings","SchemeAction",SendMessage(GetDlgItem(hwnd,IDC_SCHEMEACTION),CB_GETCURSEL,0,0) );

	SendMessage(GetDlgItem(hwnd,IDC_REFSELLIST), WM_GETTEXT, 128, (LPARAM)buf);
	for (i=0;i<NUM_REFSELLIST;i++)
		if (!strcmp(refsellist_strs[i], buf))
			HIIG("Settings","RefreshSelList",refsellist_values[i]);

	
	SendMessage(GetDlgItem(hwnd, IDT_CLAN), WM_GETTEXT, 128, (LPARAM)HashSelect(HASH_GLOBAL,"Settings","ClanId",HASH_STRING));
	SendMessage(GetDlgItem(hwnd,IDT_HTTPPROXY), WM_GETTEXT, 128, (LPARAM)buf);
	HITG("Settings","HttpProxy",buf);
	
	for (i=0;boolSettings[i].id;i++) {
		old =  HSIG("Settings",boolSettings[i].name);
     	if (SendMessage(GetDlgItem(hwnd, boolSettings[i].id), BM_GETCHECK, 0, 0) == BST_CHECKED)
			value = (!boolSettings[i].reverse);
		else value = boolSettings[i].reverse;

		if (old != value)
		{
			if (boolSettings[i].rebuildWindow == 3)
			{
				if (!value && GetLANList())
					DeleteList(GetLANList());
			}
		}
		HIIG("Settings",boolSettings[i].name,value);

		if (old != value)
		{
			if (boolSettings[i].rebuildWindow == 1)
				rebuild = 1;
			if (boolSettings[i].rebuildWindow == 2)
				ServinfoInit(NULL);

			if (boolSettings[i].rebuildWindow == 3)
			{
				if (value)
					GetLANList();
			}
			if (boolSettings[i].rebuildWindow == 4)
				SCHEME_Update();

		}
	}
	
	DestroyWindow(hwnd);
	SystrayUpdate();

	if (rebuild) 
	{
	//	CreateMainWindow(NULL);
	//	LoadWindowPosition(PROG_HWND,SW_SHOWNORMAL);
		RespawnWindow();
	}

	return 0;
}			

	


int WINAPI GENCFG_DlgProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch (msg) {
		case WM_COMMAND:
			if (LOWORD(wParam) == IDOK)
				return GENCFG_Close(hwnd);
			return 0;
		case WM_CLOSE:
			DestroyWindow(hwnd);
			return 0;
	}
	return 0;
}


void GENCFG_DialogBox(void)
{
	int i, first = -1;
	HWND hwnd;
	char tmp[20];
	char* action_strs[] = { "Connect to server","Add to favourites","Prompt user",NULL };
	hwnd = CreateDialog(PROG_HINSTANCE, MAKEINTRESOURCE(IDD_GENCFG), PROG_HWND, GENCFG_DlgProc);
		
	for (i=0;i<NUM_REFRESH;i++)
	{
		SendMessage(GetDlgItem(hwnd, IDC_REFRESH), CB_ADDSTRING, strlen(refresh_strs[i]), (LPARAM)refresh_strs[i]);
		SendMessage(GetDlgItem(hwnd, IDC_LANREFRESH), CB_ADDSTRING, strlen(refresh_strs[i]), (LPARAM)refresh_strs[i]);
	}

	for (i=0;i<NUM_REFRESH;i++)
		if ((int) HashSelect(HASH_GLOBAL,"Settings","Refresh",HASH_INT) == refresh_values[i])
			SendMessage(GetDlgItem(hwnd, IDC_REFRESH), CB_SELECTSTRING, -1, (LPARAM)refresh_strs[i]);

	for (i=0;i<NUM_TIMEOUT;i++)
	{
		SendMessage(GetDlgItem(hwnd, IDC_TIMEOUT), CB_ADDSTRING, strlen(timeout_strs[i]), (LPARAM)timeout_strs[i]);
		SendMessage(GetDlgItem(hwnd, IDC_LANTIMEOUT), CB_ADDSTRING, strlen(timeout_strs[i]), (LPARAM)timeout_strs[i]);
	}

	for (i=0;i<NUM_TIMEOUT;i++)
		if ((int) HashSelect(HASH_GLOBAL,"Settings","Timeout",HASH_INT) == timeout_values[i])
			SendMessage(GetDlgItem(hwnd, IDC_TIMEOUT), CB_SELECTSTRING, -1, (LPARAM)timeout_strs[i]);

	// FOR LAN
	for (i=0;i<NUM_REFRESH;i++)
		if ((int) HashSelect(HASH_GLOBAL,"Settings","LANRefresh",HASH_INT) == refresh_values[i])
			SendMessage(GetDlgItem(hwnd, IDC_LANREFRESH), CB_SELECTSTRING, -1, (LPARAM)refresh_strs[i]);

	for (i=0;i<NUM_TIMEOUT;i++)
		if ((int) HashSelect(HASH_GLOBAL,"Settings","LANTimeout",HASH_INT) == timeout_values[i])
			SendMessage(GetDlgItem(hwnd, IDC_LANTIMEOUT), CB_SELECTSTRING, -1, (LPARAM)timeout_strs[i]);

	// Refresh selected list
	for (i=0;i<NUM_REFSELLIST;i++)
		SendMessage(GetDlgItem(hwnd, IDC_REFSELLIST), CB_ADDSTRING, strlen(tmp), (LPARAM)refsellist_strs[i]);
	
	for (i=0;i<NUM_REFSELLIST;i++)
		if ((int) HashSelect(HASH_GLOBAL,"Settings","RefreshSelList",HASH_INT) == refsellist_values[i])
			SendMessage(GetDlgItem(hwnd, IDC_REFSELLIST), CB_SELECTSTRING, -1, (LPARAM)refsellist_strs[i]);

	SendMessage(GetDlgItem(hwnd, IDT_CLAN), WM_SETTEXT, 0, (LPARAM)HashSelect(HASH_GLOBAL,"Settings","ClanId",HASH_STRING));

	SendMessage(GetDlgItem(hwnd,IDT_HTTPPROXY), WM_SETTEXT, 0, (LPARAM)HSTG("Settings","HttpProxy") );
	/*if (HashLookupI("Settings","ShowPupServers")) 
		SendMessage(GetDlgItem(hwnd, IDC_SHOWPUPSERVERS), BM_SETCHECK, BST_CHECKED, 0);
	else EnableWindow(GetDlgItem(hwnd,IDC_REFRESHPUPSERVERS),0);*/
	
	for (i=0;boolSettings[i].id;i++)
	{
		if ((!boolSettings[i].reverse &&
			HashSelect(HASH_GLOBAL,"Settings",boolSettings[i].name,HASH_INT))
			|| (boolSettings[i].reverse && !HashSelect(HASH_GLOBAL,"Settings",boolSettings[i].name,HASH_INT)))
				SendMessage(GetDlgItem(hwnd, boolSettings[i].id), BM_SETCHECK, BST_CHECKED, 0);
	}

	for (i=0;action_strs[i];i++)
	{
		SendMessage(GetDlgItem(hwnd, IDC_SCHEMEACTION), CB_ADDSTRING, strlen(action_strs[i]), (LPARAM)action_strs[i]);
		if (HSIG("Settings","SchemeAction") == i)
			SendMessage(GetDlgItem(hwnd, IDC_SCHEMEACTION), CB_SELECTSTRING, -1, (LPARAM)action_strs[i]);
	}

	return;
}
