/*
******************************************************************
Server Query; a free games server monitoring tool
Copyright (C) 1998-2002  Nathan O'Sullivan

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
******************************************************************
*/
#include <stdio.h>
#include "types.h"
#include "resource.h"
#include "halflife\resource.h"
#include "shared.h"
#include "players.h"
#include "resolve.h"
#include "servinfo.h"


#define SVR_NUMPLAYERS (HSI(svr,"info","num_players"))
#define SVR_FRAGS(x) (HSI(svr,"frags",(HASH_PARAM)x))
#define SVR_NAME(x) (HST(svr,"name",(HASH_PARAM)x))
#define SVR_PING(x) (HSI(svr,"ping",(HASH_PARAM)x))
#define SVR_CHANGE(x) (HSI(svr,"change",(HASH_PARAM)x))

types_t* t;

SOCKET s;
char* clan_id;
unsigned long (*GetServerAddr)(char* addr);
short int (*GetServerPort)(char* addr, short otherwise);

#define DEATHS other[0]
#define PORT other[1]


void HL_Send(SOCKET s, server_t* svr, short port) 
{
	SOCKADDR_IN src_addr;
	char* str;
	char buf[128];
	FD_SET wfds, xfds;
	struct timeval tv;

	sprintf(buf,"Sending half-life packet, server %x port %d",
			svr,port);
	WRITE_DEBUG(buf);

	src_addr.sin_family = AF_INET;
	if (svr) {
		WRITE_DEBUG("Getting addr");
		src_addr.sin_addr.s_addr = GetServerAddr(svr->ipaddr);
		WRITE_DEBUG("Getting port");
		src_addr.sin_port = htons(GetServerPort(svr->ipaddr, port));
	}
	else {
		WRITE_DEBUG("Setting addr to any");
		src_addr.sin_addr.s_addr = INADDR_BROADCAST;
		WRITE_DEBUG("Getting port");
		src_addr.sin_port = htons(port);
	}
	
	str = "\xff\xff\xff\xff" "info";
	WRITE_DEBUG("Sending string 1");
	WRITE_DEBUG(str);

	FD_ZERO(&wfds); FD_ZERO(&xfds);
	FD_SET(s,&wfds); FD_SET(s,&xfds);
	tv.tv_sec = 1; tv.tv_usec = 0;
	select(0,NULL,&wfds,&xfds,&tv);

	if (FD_ISSET(s,&wfds))
		WRITE_DEBUG("write avail");
	else WRITE_DEBUG("write unavail");

	if (FD_ISSET(s,&xfds))
		WRITE_DEBUG("error occured");
	else WRITE_DEBUG("no error");
	
	sendto(s, str, strlen(str), 0, (SOCKADDR*)&src_addr, sizeof(SOCKADDR_IN));

	str = "\xff\xff\xff\xff" "rules";
	WRITE_DEBUG("Sending string 2");
	WRITE_DEBUG(str);
	sendto(s, str, strlen(str), 0, (SOCKADDR*)&src_addr, sizeof(SOCKADDR_IN));

	WRITE_DEBUG("Done, returning");

	//svr->PORT = port;

	return;
}

void HL_SendForPlayers(server_t* svr) 
{
	SOCKADDR_IN src_addr;
	char* str;
	short port = 27015;

	src_addr.sin_family = AF_INET;
	if (svr) {
		src_addr.sin_addr.s_addr = GetServerAddr(svr->ipaddr);
		src_addr.sin_port = htons(GetServerPort(svr->ipaddr, port));
	}
	else {
		src_addr.sin_addr.s_addr = INADDR_BROADCAST;
		src_addr.sin_port = htons(port);
	}
	
	str = "\xff\xff\xff\xff" "players";
	sendto(s, str, strlen(str), 0, (SOCKADDR*)&src_addr, sizeof(SOCKADDR_IN));
	
	return;
}

void HL_PrintInfo(server_t* svr, HWND mainwnd, int ignore_changes)
{
	//player_t temp_p;
	HWND temp, hwnd;
	int i, j;
	char buf[256];
	int sin = 0;
	int order[128];

	hwnd = mainwnd;

	ServinfoInit(svr);

	strcpy(buf, GetValue(svr, "hostname") );
	if (svr->timeout >= 5) {
		strcpy(buf, "UNREACHABLE");
		HashDeleteAll(svr,"Items"); // wipe all current settings so we get blanks
		HII(svr,"info","num_players", 0);
	} else if (HashCount(svr,"Items") == 0) {
		HII(svr,"info","num_players", 0);
		strcpy(buf, "CONNECTING...");
	}
	SetDlgItemText(hwnd, IDT_HOSTNAME, buf );
	ServinfoAdd("Hostname", buf);
	ServinfoAdd("Address", svr->ipaddr);

	// FIXME
	strcpy(buf,"Half-Life");
	SetDlgItemText(hwnd, IDT_VERSION,buf);
	ServinfoAdd("Game",buf);

	/*if (strcmp(GetValue(svr, "protocol"),""))
	{
		sscanf(GetValue(svr, "protocol"), "%d", &i);
		switch (i) {
			case 3:	 SetDlgItemText(hwnd, IDT_VERSION, "SIN v1.0"); break;
			case 4:	 SetDlgItemText(hwnd, IDT_VERSION, "SIN v1.01"); break;
			default: SetDlgItemText(hwnd, IDT_VERSION, "Unknown");
		}
	} else 
		SetDlgItemText(hwnd, IDT_VERSION, "");*/


	strcpy(buf, GetValue(svr, "mapname") );
	SetDlgItemText(hwnd, IDT_MAPNAME, buf );
	ServinfoAdd("Map",buf);

	if (strcmp(GetValue(svr, "maxplayers"), ""))
	{
		sprintf(buf, "%s / %s", GetValue(svr, "numplayers"), GetValue(svr, "maxplayers") );
	}
	else strcpy(buf, "");
	SetDlgItemText(hwnd, IDT_PLAYERS, buf);
	ServinfoAdd("Players",buf);

	ServinfoAdd("Ping","");
	
  if (HSIG("Settings","ExtraServInfo")) 
  {
	strcpy(buf, GetValue(svr, "mp_timelimit") );
	SetDlgItemText(hwnd, IDT_TIMELIMIT, buf );
	ServinfoAdd("Timelimit", buf);

	if (HashCount(svr,"Items") ==0 || strcmp(GetValue(svr, "mp_fraglimit"),""))
	{
		strcpy(buf, GetValue(svr, "mp_fraglimit") );
		SetDlgItemText(hwnd, IDT_FRAGLIMIT, buf);
		ServinfoAdd("Fraglimit", buf);

		

		SetDlgItemText(hwnd, IDC_CAPTURELIMIT, "Teamplay:");
		if (GetValue(svr, "mp_teamplay")[0] == '1')
			strcpy(buf, "YES" );
		else if (GetValue(svr, "mp_teamplay")[0] == '0')
			strcpy(buf, "NO" );
		else 
			strcpy(buf, "");
		SetDlgItemText(hwnd, IDT_CAPTURELIMIT, buf );
		ServinfoAdd("Teamplay",buf);
	}

	/*SetDlgItemText(hwnd, IDC_DMFLAGS, "DM Flags:" );
	strcpy(buf, GetValue(svr,"dmflags") );
	SetDlgItemText(hwnd, IDT_DMFLAGS, buf);
	ServinfoAdd("DM Flags",buf);*/

	SetDlgItemText(hwnd,IDT_MISC,"");
  }

	PlayerinfoInit(svr);
	temp = GetDlgItem(hwnd, IDL_PLAYERS);

	SendMessage(temp, LB_RESETCONTENT, 0, 0);
	
	for (i=0;i<SVR_NUMPLAYERS;i++)
		order[i] = i;

	i = 0;
	while (i < SVR_NUMPLAYERS - 1) 
	{
		if (SVR_FRAGS(order[i]) < SVR_FRAGS(order[i+1]))
		{
			j = order[i];
			order[i] = order[i+1];
			order[i+1] = j;
			i = 0;
			continue;
		}
		i++;
	}
	
	SendMessage(temp, LB_ADDSTRING, 0, (LPARAM)"\tKills\tPing\tName");
	
	for (i=0;i<SVR_NUMPLAYERS;i++)
	{
		strcpy(buf,"");

		if (strcmp(clan_id,""))
		{
			for (j=0;j<(int)(strlen(SVR_NAME(order[i])) - strlen(clan_id));j++)
				if (!memcmp(SVR_NAME(order[i])+j, clan_id, strlen(clan_id)))
					strcat(buf,"*");
		}
		
		strcat(buf, "\t");

		if (!ignore_changes && SVR_CHANGE(order[i]))
			sprintf(buf + strlen(buf), "%d (%c%d)\t?\t%s", 
			SVR_FRAGS(order[i]),  
			SVR_CHANGE(order[i]) > 0 ? '+' : ' ',
			SVR_CHANGE(order[i]),
			SVR_NAME(order[i])
		);
		
		else sprintf(buf + strlen(buf), "%d\t?\t%s", 
			SVR_FRAGS(order[i]),
			SVR_NAME(order[i])
			);
		
		SendMessage(temp, LB_ADDSTRING, 0, (LPARAM)buf);

		PlayerinfoAdd(SVR_NAME(order[i]),SVR_FRAGS(order[i]),SVR_CHANGE(order[i]),-1000);
	}
	return;
}

char* HL_Hostname(server_t* svr)
{
	return GetValue(svr, "hostname");
}

int GetString(char* ptr, char* buf, int max)
{
	int i = 0;
	while (i < max)
	{
		*buf = *(ptr++);
		i++;
		if (*buf == 0)
			break;
		else buf++;
		
	}
	if (i == max) {
		buf[i - 1] = 0;
		while (i < 512) {
			if (*(ptr++) == 0)
				break;
		}
	}
	return i;
}
	
void DecodeInfoString(server_t* svr, char* str)
{
	char* ptr = str;
	char buf[256];

	HashDeleteAll(svr,"Items");			// clean out all the crap

	ptr+= GetString(ptr,buf, 250);  // - ip we already know

	ptr+= GetString(ptr,buf, 250);
	HashInsert(svr,"Items", "hostname", buf, HASH_STRING);
	ptr+= GetString(ptr,buf, 250);
	HashInsert(svr,"Items", "mapname", buf, HASH_STRING);
	
	ptr+= GetString(ptr,buf, 250); // mod dir
	HashInsert(svr,"items","game",buf,HASH_STRING);
	ptr+= GetString(ptr,buf, 250); // mod name
	HashInsert(svr,"items","gamename",buf,HASH_STRING);
	sprintf(buf, "%d", (unsigned char) *(ptr++));
	HashInsert(svr,"Items", "numplayers", buf,HASH_STRING);
	sprintf(buf, "%d", (unsigned char) *(ptr++));
	HashInsert(svr,"Items", "maxplayers", buf,HASH_STRING);

	HashInsertInt(svr,"Items","norules",1);

	HL_SendForPlayers(svr);
	
	return;
}

void DecodeRulesString(server_t* svr, char* str, char* maxstr)
{
	char* ptr = str;
	char key[256];
	char value[256];
	int number, i;
	


	number = *(char*)ptr;
	ptr+=2;

	
	for (i=0;i<number;i++)
	{
		ptr += GetString(ptr,key, 30);  // setting key
		ptr += GetString(ptr,value, 250);  // setting value
		HashInsert(svr,"items",key, value,HASH_STRING);
	}		

	if (HSI(svr,"items","reserve_slots"))
		HII(svr,"items","maxplayers", HSI(svr,"items","maxplayers") - 
		HSI(svr,"items","reserve_slots"));

	
	return;
}

void DecodePlayersString(server_t* svr, char* str, char* maxstr)
{
	char* ptr = str;
	char buf[256];
	
	HII(svr,"info","num_players",0);
	ptr++; // num_players

	while (ptr < maxstr)
	{
		ptr+=1; // player number
		ptr+= GetString(ptr,buf, 32);  // player name
		HIT(svr,"name",(HASH_PARAM)SVR_NUMPLAYERS,buf);
		HII(svr,"frags",(HASH_PARAM)SVR_NUMPLAYERS,*(short*)ptr);
		ptr+=2;
		//svr->players[svr->num_players].DEATHS = *(short*)ptr;
		ptr+=2;
		//svr->players[svr->num_players].ping = *(short*)ptr;
		ptr+=4;//unknown
		HII(svr,"info","num_players",SVR_NUMPLAYERS + 1);
	}
	CleanRecords(svr);
	return;
}

int HL_Receive(server_t* svr, char* buf, int len)
{
	char* ptr;
	
	ptr = buf;

	if (memcmp(ptr, "\xfe\xff\xff\xff",4) == 0)
	{
		if (HashSelectInt(svr,"items","norules") == 1)
			HashDelete(svr,"items","norules");

		ptr += 8;
		if (memcmp(ptr, "\xff\xff\xff\xff",4) == 0)
			ptr += 5;
		DecodeRulesString(svr,ptr,buf+len);

		return 0;
	}
	if (memcmp(ptr, "\xff\xff\xff\xff",4))
		return 0;

	ptr+=4;
	if (*ptr == 'C') 
		DecodeInfoString(svr,ptr+1);

	else if (*ptr == 'D')
		DecodePlayersString(svr,ptr+1,buf+len);
		
	else if (*ptr == 'E')
	{
		if (HashSelectInt(svr,"items","norules") == 1)
			HashDelete(svr,"items","norules");
		DecodeRulesString(svr,ptr+1,buf+len);
	}

	// Update available information
	HashInsertText(svr,"stats","map",
		HashSelectText(svr,"items","mapname"));
	HashInsertText(svr,"stats","numplayers",
		HashSelectText(svr,"items","numplayers") );
	HashInsertText(svr,"stats","maxplayers",
		HashSelectText(svr,"items","maxplayers") );

	return 1; // update displayed info
}

char* HL_Cmdline(server_t* svr)
{
	static char string[256];
	sprintf(string, "-console ");
	if (HashExistsKey(svr,"items","game") != (HASH_PARAM)-1)
		sprintf(string+strlen(string),"-game %s ",HashSelectText(svr,"items","game") );

	// If there's a password, set it
	if (strcmp(HashSelect(HASH_GLOBAL,"Status","Password",HASH_STRING),"") != 0)
		sprintf(string+strlen(string)," +password \"%s\"",HashSelect(HASH_GLOBAL,"Status","Password",HASH_STRING));

	sprintf (string+strlen(string),"+connect %s", svr->ipaddr);
	return string;
}


int HL_IsFull(server_t* svr)
{
	if (atoi( GetValue(svr,"numplayers") ) && atoi( GetValue(svr,"maxplayers") ) &&
			atoi ( GetValue(svr,"numplayers") ) == atoi( GetValue(svr, "maxplayers") ) )
		return 1;
	return 0;
}

int __stdcall Init(shared_t* info)
{
	
	GetServerAddr = info->GetServerAddr;
	GetServerPort = info->GetServerPort;

	HashInit(info->hashes,info->numHashes);
	s = HashSelectInt(HASH_GLOBAL,"Global","Socket");
	clan_id = HashSelectText(HASH_GLOBAL,"Settings","ClanId");
	


	t = info->types[0];
	//t->type = 4;
	t->protocol = 1;
	t->port = 27015;
	strcpy(t->name, "Half-life");
	strcpy(t->path, "C:\\Half-life\\hl.exe");
	strcpy(t->qstat_type, "HLS");
	strcpy(t->gamespy_type, "hlprot");
	strcpy(t->mod_var,"game");
	t->receive = HL_Receive;
	t->print_info = HL_PrintInfo;
	t->send = HL_Send;
	t->hostname = HL_Hostname;
	//t->cmdline = HL_Cmdline;
	t->isfull = HL_IsFull;


	//HashInsertPtr(t,"funcs","isfull",HL_IsFull);

	/*(void*)t->iImage = LoadBitmap(
		HashSelectPtr(HASH_GLOBAL,"Status","Hinstance"),
		MAKEINTRESOURCE(IDB_GAMEICON)
	);*/

	return 1;
}

