/*
******************************************************************
Server Query; a free games server monitoring tool
Copyright (C) 1998-2002  Nathan O'Sullivan

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
******************************************************************
*/
#include <stdio.h>
#include "resource.h"
#include "quake\resource.h"
#include "types.h"
#include "shared.h"
#include "quakex.h"
#include "servinfo.h"

#define SVR_NUMPLAYERS (HSI(svr,"info","num_players"))
#define SVR_FRAGS(x) (HSI(svr,"frags",(HASH_PARAM)x))
#define SVR_NAME(x) (HST(svr,"name",(HASH_PARAM)x))
#define SVR_PING(x) (HSI(svr,"ping",(HASH_PARAM)x))
#define SVR_CHANGE(x) (HSI(svr,"change",(HASH_PARAM)x))

SOCKET s;
char* clan_id;
unsigned long (*GetServerAddr)(char* addr);
short int (*GetServerPort)(char* addr, short otherwise);

unsigned short switchBytes(unsigned short num)
{
	unsigned short i, j;
	i = num & 0xff;
	j = num >> 8;
	return (i << 8) + j;
}

typedef struct {
	unsigned short control;
	unsigned short len;
	unsigned char opcode;
	char data[256];
} QPACKET;


void QUAKE_SendForPlayers(server_t* svr) 
{
	SOCKADDR_IN src_addr;
    QPACKET pkt;

	src_addr.sin_family = AF_INET;
	if (svr) {
		src_addr.sin_addr.s_addr = GetServerAddr(svr->ipaddr);
		src_addr.sin_port = htons(GetServerPort(svr->ipaddr, 26000)); // FIXME
	}
	else {
		src_addr.sin_addr.s_addr = INADDR_BROADCAST;
		src_addr.sin_port = htons(26000); //FIXME
	}
	
	pkt.control = switchBytes(0x8000);
	pkt.opcode = 0x03;
	pkt.data[0] = svr->other[5];
	pkt.len = switchBytes(6);
	
	sendto(s, (char*)&pkt, 6, 0, (SOCKADDR*)&src_addr, sizeof(SOCKADDR_IN));
	
	return;
}
void QUAKE_SendForRules(server_t* svr) 
{
	SOCKADDR_IN src_addr;
    QPACKET pkt;

	src_addr.sin_family = AF_INET;
	if (svr) {
		src_addr.sin_addr.s_addr = GetServerAddr(svr->ipaddr);
		src_addr.sin_port = htons(GetServerPort(svr->ipaddr, 26000)); // FIXME
	}
	else {
		src_addr.sin_addr.s_addr = INADDR_BROADCAST;
		src_addr.sin_port = htons(26000); //FIXME
	}
	
	pkt.control = switchBytes(0x8000);
	pkt.opcode = 0x04;
	strcpy(pkt.data,HashSelectText(svr,"items","nextrule"));
	pkt.len = switchBytes((unsigned short)(6 + strlen(pkt.data)));
	
	sendto(s, (char*)&pkt, 6 + strlen(pkt.data), 0, (SOCKADDR*)&src_addr, sizeof(SOCKADDR_IN));
	
	return;
}
void QUAKE_Send(SOCKET s, server_t* svr, short port) 
{
	SOCKADDR_IN src_addr;
    QPACKET pkt;

	src_addr.sin_family = AF_INET;
	if (svr) {
		src_addr.sin_addr.s_addr = GetServerAddr(svr->ipaddr);
		src_addr.sin_port = htons(GetServerPort(svr->ipaddr, port));
	}
	else {
		src_addr.sin_addr.s_addr = INADDR_BROADCAST;
		src_addr.sin_port = htons(port);
	}
	
	pkt.control = switchBytes(0x8000); 
	pkt.len = switchBytes(12);
	pkt.opcode = 0x02;
	memcpy(pkt.data,"QUAKE\000\003",7);
	
	sendto(s, (char*)&pkt, 12, 0, (SOCKADDR*)&src_addr, sizeof(SOCKADDR_IN));
	
	return;
}


void QUAKE_PrintInfo(server_t* svr, HWND mainwnd, int ignore_changes)
{
	HWND temp, hwnd;
	int i, j;
	char buf[128];
	unsigned char* ptr;
	//player_t temp_p;

	hwnd = mainwnd;

	ServinfoInit(svr);

	strcpy(buf, GetValue(svr, "hostname") );
	if (svr->timeout >= 5) {
		strcpy(buf, "UNREACHABLE");
		HashDeleteAll(svr,"items"); // wipe all current settings so we get blanks
		HII(svr,"info","num_players",0);
	} else if (HashCount(svr,"items") == 0) {
		HII(svr,"info","num_players",0);
		strcpy(buf, "CONNECTING...");
	}
	SetDlgItemText(hwnd, IDT_HOSTNAME, buf );
	ServinfoAdd("Hostname", buf);
	ServinfoAdd("Address", svr->ipaddr);
	
	strcpy(buf, "Quake");
	SetDlgItemText(hwnd, IDT_VERSION, buf); 
	ServinfoAdd("Game", buf);

	strcpy(buf, GetValue(svr, "mapname") );
	if (!strcmp( GetValue(svr, "mapname"), "") || GetValue(svr, "mapname") == NULL)
		strcpy(buf, GetValue(svr, "map") );
	SetDlgItemText(hwnd, IDT_MAPNAME, buf );
	ServinfoAdd("Map",buf);
	
	if (strcmp(GetValue(svr, "maxplayers"), ""))
	{
		sprintf(buf, "%s / %s", GetValue(svr,"numplayers"), GetValue(svr, "maxplayers") );
	}
	else strcpy(buf, "");
	SetDlgItemText(hwnd, IDT_PLAYERS, buf);
	ServinfoAdd("Players",buf);

	ServinfoAdd("Ping", "");

  if (HSIG("Settings","ExtraServInfo"))
  {
	
	strcpy(buf, GetValue(svr, "fraglimit") );
	SetDlgItemText(hwnd, IDT_FRAGLIMIT, buf );
	ServinfoAdd("Fraglimit", buf);

	strcpy(buf, GetValue(svr, "timelimit") );
	SetDlgItemText(hwnd, IDT_TIMELIMIT, buf );
	ServinfoAdd("Timelimit", buf);

	SetDlgItemText(hwnd, IDC_CAPTURELIMIT, "Deathmatch:" );
	strcpy(buf, GetValue(svr,"deathmatch") );
	SetDlgItemText(hwnd, IDT_CAPTURELIMIT, buf );
	ServinfoAdd("Deathmatch", buf);

	
	SetDlgItemText(hwnd, IDC_DMFLAGS, "Teamplay:" );
	strcpy(buf, GetValue(svr, "teamplay") );
	SetDlgItemText(hwnd, IDT_DMFLAGS, buf);
	ServinfoAdd("Teamplay", buf);

	SetDlgItemText(hwnd,IDT_MISC,"");
  }
	temp = GetDlgItem(hwnd, IDL_PLAYERS);

	SendMessage(temp, LB_RESETCONTENT, 0, 0);
	SendMessage(temp, LB_ADDSTRING, 0, (LPARAM)"\tFrags\tPing\tName");
	
	/*if (svr->num_players == atoi(GetValue(svr,"numplayers")))
	{
		i = 0;
		while (i < svr->num_players - 1) 
		{
			if (svr->players[i].frags < svr->players[i+1].frags)
			{
				temp_p = svr->players[i];
				svr->players[i] = svr->players[i+1];
				svr->players[i+1] = temp_p;
				i =	0;
				continue;
			}
			i++;
		}
	}*/

	for (i=0;i<SVR_NUMPLAYERS;i++)
	{
		ptr = SVR_NAME(i);
		for (j=0;j<(int) strlen(ptr);j++)
			if ((ptr[j] >= 'A' + 128 && ptr[j]<='Z'+128) ||
				(ptr[j] >= 'a' + 128 && ptr[j]<='z'+128))
				ptr[j]-=128;
	}


	PlayerinfoInit(svr);

	for (i=0;i<SVR_NUMPLAYERS;i++)
	{
		strcpy(buf,"\t");

		if (strcmp(clan_id,""))
		{
			for (j=0;j<(int)(strlen(SVR_NAME(i)) - strlen(clan_id));j++)
				if (!memcmp(SVR_NAME(i)+j, clan_id, strlen(clan_id)))
					strcpy(buf,"*\t");
		}
		
		if (!ignore_changes && SVR_CHANGE(i))
			sprintf(buf + strlen(buf), "%d (%c%d)\t%d\t%s", 
				SVR_FRAGS(i),
				SVR_CHANGE(i) > 0 ? '+' : ' ',
				SVR_CHANGE(i),
				SVR_PING(i),
				SVR_NAME(i)
			);
		else sprintf(buf + strlen(buf), "%d\t%d\t%s", 
				SVR_FRAGS(i), 
				SVR_PING(i),
				SVR_NAME(i)
			);
		SendMessage(temp, LB_ADDSTRING, 0, (LPARAM)buf);
		PlayerinfoAdd(SVR_NAME(i),SVR_FRAGS(i),SVR_CHANGE(i),SVR_PING(i));
	}
	return;
}

char* QUAKE_Hostname(server_t* svr)
{
	return GetValue(svr, "hostname");
}


int GetString(char* ptr, char* buf, int max)
{
	int i = 0;
	while (i < max)
	{
		*buf = *(ptr++);
		i++;
		if (*buf == 0)
			break;
		else buf++;
		
	}
	if (i == max) {
		buf[i - 1] = 0;
		while (i < 512) {
			if (*(ptr++) == 0)
				break;
		}
	}
	return i;
}
int QUAKE_Receive(server_t* svr, char* buf, int len)
{
	QPACKET* pkt = (QPACKET*)buf;
	char* str = pkt->data;
	char temp[128], value[128];
	int num;

	buf[len] = 0;
	
/*	FILE* fp;
	fp = fopen("debug.txt","ab");
	fwrite(buf,1,len,fp);
	fclose(fp);*/

	if (switchBytes(pkt->control) != 0x8000)
		return 0; // bogus packet
	
	if (switchBytes(pkt->len) != len)
		return 0; // bugus packet
	
	if (pkt->opcode == 0x83)
	{ // server info

		str += GetString(str,temp,128); // SERVER IP
		str += GetString(str,temp,128); // HOSTNAME
		HashInsertText(svr,"items","hostname", temp);
		str += GetString(str,temp,128); // MAPNAME
		HashInsertText(svr,"items","mapname", temp);
		sprintf(temp, "%d",(unsigned char*)str[0]); str++;
		HashInsertText(svr,"items","numplayers", temp);
		sprintf(temp, "%d",(unsigned char*)str[0]); str++;
		HashInsertText(svr,"items","maxplayers", temp);
		
		HashInsertText(svr,"items", "nextrule", "");
		svr->other[5] = 0;
		//svr->num_players = 0;
		//UHashClear(svr->items);

		QUAKE_SendForRules(svr);
		QUAKE_SendForPlayers(svr);
	} 
	else if (pkt->opcode == 0x84)
	{	// player info
		num = str[0];
		str++;
		str += GetString(str,temp,128); // NAME
		HIT(svr,"name",(HASH_PARAM)num,temp);

		str += 4;
		HII(svr,"frags",(HASH_PARAM)num,((unsigned int*)str)[0] );

		if (num == SVR_NUMPLAYERS)
			HII(svr,"info","num_players", SVR_NUMPLAYERS + 1);

		if (num == HashSelectInt(svr,"items","numplayers") )
		{
			HII(svr,"info","num_players",HSI(svr,"items","numplayers") );
			return 1;
		}

		svr->other[5]++;
		QUAKE_SendForPlayers(svr);
		return 0;
	}
	else if (pkt->opcode == 0x85)
	{	// rule info
		if (! pkt->data[0])
			return 1;
		str += GetString(str,temp,128); // CVAR
		str += GetString(str,value,128); // VALUE
		HashInsertText(svr,"items",temp,value);
		HashInsertText(svr,"items","nextrule",temp);
		QUAKE_SendForRules(svr);
		return 0;
	}

	// Update available information
	HashInsertText(svr,"stats","map",
		HashSelectText(svr,"items","map"));
	HashInsertText(svr,"stats","numplayers",
		HashSelectText(svr,"items","numplayers") );
	HashInsertText(svr,"stats","maxplayers",
		HashSelectText(svr,"items","maxplayers") );

	return 1; // update displayed info
}

int QUAKE_IsFull(server_t* svr)
{
	if (atoi( GetValue(svr,"numplayers") ) && atoi( GetValue(svr, "maxplayers") ) && 
			atoi ( GetValue(svr,"numplayers") ) == atoi( GetValue(svr, "maxplayers") ) )
		return 1;
	return 0;
}

int __stdcall Init(shared_t* info)
{
	types_t* t;

	
	GetServerAddr = info->GetServerAddr;
	GetServerPort = info->GetServerPort;

	HashInit(info->hashes,info->numHashes);
	s = HashSelectInt(HASH_GLOBAL,"Global","Socket");
	clan_id = HashSelectText(HASH_GLOBAL,"Settings","ClanId");
	

	t = info->types[0];
	//t->type = 8;
	t->port = 26000;
	strcpy(t->name, "Quake");
	strcpy(t->path, "C:\\Quake\\glquake.exe");
	strcpy(t->qstat_type,"QS");
	strcpy(t->gamespy_type,"quake");
	strcpy(t->mod_var,"game");
	t->receive = QUAKE_Receive;
	t->print_info = QUAKE_PrintInfo;
	t->send = QUAKE_Send;
	t->hostname = QUAKE_Hostname;
	t->isfull = QUAKE_IsFull;

	//HashInsertPtr(t,"funcs","isfull",QUAKE_IsFull);

	
	(void*)t->iImage = LoadBitmap(
		HashSelectPtr(HASH_GLOBAL,"Status","Hinstance"),
		MAKEINTRESOURCE(IDB_GAMEICON)
	);
		
	return 1;
}

