#include <stdio.h>
#include "resource.h"
#include "quake2\resource.h"
#include "types.h"
#include "shared.h"
#include <time.h>
#include "quakex.h"
#include "resolve.h"
#include "servinfo.h"

#define TYPE_QUAKE2 1
#define TYPE_SIN 2
#define TYPE_KINGPIN 9
#define TYPE_SOF 12
#define TYPE_DAIK 13

#define SVR_NUMPLAYERS (HSI(svr,"info","num_players"))
#define SVR_FRAGS(x) (HSI(svr,"frags",(HASH_PARAM)x))
#define SVR_NAME(x) (HST(svr,"name",(HASH_PARAM)x))
#define SVR_PING(x) (HSI(svr,"ping",(HASH_PARAM)x))
#define SVR_CHANGE(x) (HSI(svr,"change",(HASH_PARAM)x))
//quake2 only
#define SVR_TEAM(x) (HSI(svr,"team",(HASH_PARAM)x))
#define SVR_FLAG(x) (HSI(svr,"flag",(HASH_PARAM)x))

// Obtained from the executable
SOCKET s;
char* clan_id;
unsigned long (*GetServerAddr)(char* addr);
short int (*GetServerPort)(char* addr, short otherwise);
HWND mainhwnd;


void QUAKE2_Send(SOCKET s, server_t* svr, short port) 
{
	SOCKADDR_IN src_addr;

	src_addr.sin_family = AF_INET;
	if (svr) {
		src_addr.sin_addr.s_addr = GetServerAddr(svr->ipaddr);
		src_addr.sin_port = htons(GetServerPort(svr->ipaddr, port));
	}
	else {
		src_addr.sin_addr.s_addr = INADDR_BROADCAST;
		src_addr.sin_port = htons(port);
	}
	
	sendto(s, "\xff\xff\xff\xffstatus", 11, 0, (SOCKADDR*)&src_addr, sizeof(SOCKADDR_IN));
	
	if (svr && strcmp(svr->gametype->qstat_type,"DKS") == 0) // svr && svr->type == TYPE_DAIK)
	{
		//HashDeleteAll(svr,"items");
		sendto(s,"\xff\xff\xff\xffinfo 31",12,0,(SOCKADDR*)&src_addr,sizeof(SOCKADDR_IN));
		sendto(s,"\xff\xff\xff\xffinfo 32",12,0,(SOCKADDR*)&src_addr,sizeof(SOCKADDR_IN));
	}


	
	return;
}

void QUAKE2_PrintInfo(server_t* svr, HWND mainwnd, int ignore_changes)
{
//	player_t temp_p;
	HWND temp, hwnd;
	int i, j, teams = 0;
	int caps[2];
	char buf[128];
	char timerem[128];

	int sin = 0;
	int num_players;
	int order[128];

	hwnd = mainwnd;

	mainhwnd = mainwnd;

	ServinfoInit(svr);

	sprintf(buf, GetValue(svr,"hostname") );

	if (svr->timeout >= 5) {
		strcpy(buf, "UNREACHABLE");
		HashDeleteAll(svr,"items");		// wipe all current settings so we get blanks
		HII(svr,"info","num_players",0);
	} else if (HashCount(svr,"items") == 0) {
		HII(svr,"info","num_players",0);
		strcpy(buf,"CONNECTING...");
	}
	SetDlgItemText(hwnd, IDT_HOSTNAME, buf );
	ServinfoAdd("Hostname", buf);
	ServinfoAdd("Address", svr->ipaddr);

	strcpy(buf, "");

	if (strcmp(svr->gametype->qstat_type,"Q2S") == 0) //  svr->type == TYPE_QUAKE2) {
	{	
		if (strcmp(GetValue(svr, "protocol"),""))
		{
			sscanf(GetValue(svr, "protocol"), "%d", &i);
			switch (i) {
				case 26: strcpy(buf, "Quake 2  (v3.05)"); break;
				case 27: strcpy(buf, "Quake 2  (v3.08)"); break;
				case 28: strcpy(buf, "Quake 2  (v3.09)"); break;
				case 30: strcpy(buf, "Quake 2  (v3.10)"); break;
				case 31: strcpy(buf, "Quake 2  (v3.14)"); break;
				case 32: strcpy(buf, "Quake 2  (v3.15)"); break;
				case 33: strcpy(buf, "Quake 2  (v3.17)"); break;
				case 34: strcpy(buf, "Quake 2  (v3.20)"); break;
				default: strcpy(buf, "Quake 2");
			}
		}
		else strcpy(buf,"Quake2");
	}
	else if (strcmp(svr->gametype->qstat_type,"SFS") == 0) { // svr->type == TYPE_SOF) {
		sprintf(buf,"SOF ");
		sscanf(GetValue(svr,"version"),"%s",buf+strlen(buf));
	}
	else if (strcmp(svr->gametype->qstat_type,"KPS") == 0) { // svr->type == TYPE_KINGPIN) {
		if (strcmp(GetValue(svr, "protocol"),""))
		{
			sscanf(GetValue(svr, "protocol"), "%d", &i);
			switch (i) {
				case 32: strcpy(buf, "Kingpin v1.0+"); break;
				default: strcpy(buf, "Kingpin");
			}
		}
		else strcpy(buf,"Kingpin");
	}
	
	else if (strcmp(svr->gametype->qstat_type,"SNS") == 0) { // svr->type == TYPE_SIN) {
		if (strcmp(GetValue(svr, "protocol"),""))
		{
			sscanf(GetValue(svr, "protocol"), "%d", &i);
			switch (i) {
				case 3:	 strcpy(buf, "SIN  (v1.0)"); break;
				case 4:	 strcpy(buf, "SIN  (v1.01)"); break;
				default: strcpy(buf, "SIN");
			}
		}
		else strcpy(buf, "SIN");
		if (strcmp(GetValue(svr,"map"),""))
			strcpy(buf, "SIN Demo (v0.9)");
	}
	else if (strcmp(svr->gametype->qstat_type,"DKS") == 0)  // svr->type == TYPE_DAIK) 
		strcpy(buf, "Daikatana");
	else 
		strcpy(buf, "");
	SetDlgItemText(hwnd,IDT_VERSION,buf);
	ServinfoAdd("Game",buf);

	strcpy(buf, GetValue(svr, "mapname") );
	if (!strcmp( GetValue(svr, "mapname"), "") || GetValue(svr, "mapname") == NULL)
	{
		strcpy(buf, GetValue(svr, "map") );
	
	}
	SetDlgItemText(hwnd, IDT_MAPNAME, buf );
	ServinfoAdd("Map",buf);
	
	if (strcmp(GetValue(svr, "maxclients"), ""))
	{
		sprintf(buf, "%d / %s", HSI(svr,"info","num_players"), GetValue(svr, "maxclients") );
	}
	else strcpy(buf, "");
	SetDlgItemText(hwnd, IDT_PLAYERS, buf);
	ServinfoAdd("Players",buf);

	// == Set ping ==
	ServinfoAdd("Ping","");

  if (HSIG("Settings","ExtraServInfo"))
  {
	strcpy(buf, GetValue(svr, "fraglimit") );
	SetDlgItemText(hwnd, IDT_FRAGLIMIT, buf);
	ServinfoAdd("Fraglimit",buf);

	strcpy(buf, GetValue(svr, "timelimit") );
	if (GetValue(svr, "gamestats")[0])
	{
		sscanf(GetValue(svr,"gamestats"), "%s", timerem);
		sprintf(buf, "%s (%s remaining)",GetValue(svr, "timelimit"),timerem);
	}
	SetDlgItemText(hwnd, IDT_TIMELIMIT, buf);
	ServinfoAdd("Timelimit",buf);

	SetDlgItemText(hwnd, IDC_CAPTURELIMIT, "Capturelimit:");
	strcpy(buf, GetValue(svr, "capturelimit") );
	SetDlgItemText(hwnd, IDT_CAPTURELIMIT, buf );
	ServinfoAdd("Capturelimit",buf);

	/*SetDlgItemText(hwnd, IDC_DMFLAGS, "DM Flags:" );
	strcpy(buf, GetValue(svr,"dmflags") );
	SetDlgItemText(hwnd, IDT_DMFLAGS, buf);
	ServinfoAdd("DM Flags", buf);*/

	if (GetValue(svr, "gamestats")[0])
	{
		sscanf(GetValue(svr,"gamestats"), "%*s %d/%*d %d/%*d", caps,caps+1);
		if (caps[0] == caps[1])
			sprintf(buf,"Drawn at %d caps apiece.",caps[0]);
		else if (caps[1] > caps[0])
			sprintf(buf,"BLUE leads RED, %d caps to %d",caps[1],caps[0]);
		else sprintf(buf,"RED leads BLUE, %d caps to %d",caps[0],caps[1]);

		ServinfoAdd("Score", buf);
	}
  }

	temp = GetDlgItem(hwnd, IDL_PLAYERS);

	SendMessage(temp, LB_RESETCONTENT, 0, 0);
	
	num_players = HSI(svr,"info","num_players");

	for (i=0;i<num_players;i++)
		order[i] = i;

	i = 0;
	while (i < num_players - 1) 
	{
		if (SVR_FRAGS(order[i]) < SVR_FRAGS(order[i+1]))
		{
			j = order[i];
			order[i] = order[i+1];
			order[i+1] = j;
			i = 0;
			continue;
		}
		i++;
	}
	

	PlayerinfoInit(svr);

	SendMessage(temp, LB_ADDSTRING, 0, (LPARAM)"\tFrags\tPing\tName");
		
	if (!teams)
	{
		for (i=0;i<num_players;i++)
		{
			strcpy(buf,"");

			if (SVR_TEAM(order[i]) == 1)
				strcat(buf, "R");
			else if (SVR_TEAM(order[i]) == 2)
				strcat(buf, "B");
			if (SVR_FLAG(order[i]))
				strcat(buf, "F");

			if (strcmp(clan_id,""))
			{
				for (j=0;j<(int)(strlen(SVR_NAME(order[i])) - strlen(clan_id));j++)
					if (!memcmp(SVR_NAME(order[i])+j, clan_id, strlen(clan_id)))
						strcat(buf,"*");
			}
		
			strcat(buf, "\t");

			if (!ignore_changes && SVR_CHANGE(order[i]))
				sprintf(buf + strlen(buf), "%d (%c%d)\t%d\t%s", 
					SVR_FRAGS(order[i]),
					SVR_CHANGE(order[i]) > 0 ? '+' : ' ', 
					SVR_CHANGE(order[i]),
					SVR_PING(order[i]),
					SVR_NAME(order[i])
				);
			else 
				sprintf(buf + strlen(buf), "%d\t%d\t%s", 
					SVR_FRAGS(order[i]),
					SVR_PING(order[i]),
					SVR_NAME(order[i])
				);
			SendMessage(temp, LB_ADDSTRING, 0, (LPARAM)buf);
			PlayerinfoAdd(SVR_NAME(order[i]),SVR_FRAGS(order[i]),SVR_CHANGE(order[i]),SVR_PING(order[i]));

			
		}
	}

	if (GetValue(svr, "gamestats")[0])
	{
		sscanf(GetValue(svr,"gamestats"), "%*s %d/%*d %d/%*d", caps,caps+1);
		if (caps[0] == caps[1])
			sprintf(buf,"Drawn at %d caps apiece.",caps[0]);
		else if (caps[1] > caps[0])
			sprintf(buf,"BLUE leads RED, %d caps to %d.",caps[1],caps[0]);
		else sprintf(buf,"RED leads BLUE, %d caps to %d.",caps[0],caps[1]);

		SetDlgItemText(hwnd,IDT_MISC,buf);
		
		SendMessage(temp, LB_ADDSTRING, 0, (LPARAM)"");
		SendMessage(temp, LB_ADDSTRING, 0, (LPARAM)buf);

	}
	else SetDlgItemText(hwnd,IDT_MISC,"");

	/*
	else while (teams != 4) {
		if (teams == 1) {
			SendMessage(temp, LB_ADDSTRING, 0, (LPARAM)"");
			sprintf(buf,"\tRED TEAM - %d caps, %d frags",caps[0],frags[0]);
			SendMessage(temp, LB_ADDSTRING, 0, (LPARAM)buf);
		}
		else if (teams == 2) {
			SendMessage(temp, LB_ADDSTRING, 0, (LPARAM)"");
			sprintf(buf,"\tBLUE TEAM - %d caps, %d frags",caps[1],frags[1]);
			SendMessage(temp, LB_ADDSTRING, 0, (LPARAM)buf);
		}			
		if (teams<= 2) {
			for (i=0;i<num_players;i++)
			{
				strcpy(buf,"");
	
				if (SVR_TEAM(order[i]) != teams)
					continue;
				if (SVR_FLAG(order[i]))
					strcat(buf, "F");
	
				if (strcmp(clan_id,""))
				{
					for (j=0;j<(int)(strlen(SVR_NAME(order[i])) - strlen(clan_id));j++)
						if (!memcmp(SVR_NAME(order[i])+j, clan_id, strlen(clan_id)))
							strcat(buf,"*");
				}
		
				strcat(buf, "\t");
	
				if (!ignore_changes && SVR_CHANGE(order[i]))
					sprintf(buf + strlen(buf), "%d (%c%d)\t%d\t%s", 
						SVR_FRAGS(order[i]),
						SVR_CHANGE(order[i]) > 0 ? '+' : ' ',
						SVR_CHANGE(order[i]),
						SVR_PING(order[i]),
						SVR_NAME(order[i])
					);
				else 
					sprintf(buf + strlen(buf), "%d\t%d\t%s", 
						SVR_FRAGS(order[i]),
						SVR_PING(order[i]),
						SVR_NAME(order[i])
					);
					
				SendMessage(temp, LB_ADDSTRING, 0, (LPARAM)buf);
			}
		}
		else  {
			j = 0; vals[0] = 0;
			for (i=0;i<num_players;i++)
			{
				if (j == 0)
					strcpy(buf,""); 
	
				if (SVR_TEAM(order[i]) != 0)
					continue;
				
				sprintf(buf+strlen(buf), "%s%s",(j==1?"\t\t":"\t"),
					SVR_NAME(order[i])
				);
				j++;
				if (j == 2) {
					if (!vals[0])
					{
						SendMessage(temp, LB_ADDSTRING, 0, (LPARAM)"");
						SendMessage(temp, LB_ADDSTRING, 0,(LPARAM)"\tSPECTATORS");
						vals[0] = 1;
					}
		
					SendMessage(temp, LB_ADDSTRING, 0, (LPARAM)buf);
					j = 0;
				}
			}
			if (j == 1) {
				if (!vals[0])
				{
					SendMessage(temp, LB_ADDSTRING, 0, (LPARAM)"");
					SendMessage(temp, LB_ADDSTRING, 0,(LPARAM)"\tSPECTATORS");
					vals[0] = 1;
				}
				SendMessage(temp, LB_ADDSTRING, 0, (LPARAM)buf);
				return;
			}
		}
		teams++;
	}
	*/
	#ifdef ULTRA_SQ
	AdminStuff(svr);
	#endif

		return;
}

char* QUAKE2_Hostname(server_t* svr)
{
	return GetValue(svr, "hostname");
}


int DAIK_DecodeInfoString(server_t* svr, char* buf, int len)
{
	char* ptr;
	char* tmp;
	
	strtok(buf,"\n");
	ptr = strtok(NULL,"\n");
		
	if (strstr(ptr,"wrong version") != NULL)
		return 0;

	tmp = strtok(ptr,"/");
	HIT(svr,"items","hostname",tmp);
  
	tmp = strtok(NULL,"/");
	HIT(svr,"items","mapname",tmp);

	tmp = strtok(NULL,"/");
	//HIT(svr,"items","numplayers",tmp);

	tmp = strtok(NULL,"/");
	HIT(svr,"items","maxclients",tmp);

	tmp = strtok(NULL,"/");
	HIT(svr,"items","game",tmp);

	return 1;
}





int QUAKE2_Receive(server_t* svr, char* buf, int len)
{
	buf[len] = 0;


	if (memcmp(buf,"\xff\xff\xff\xffinfo",8) != 0)
		DecodeString(svr, buf, len);
	else if (!DAIK_DecodeInfoString(svr,buf,len))
		return 0;

	
	// Update available information
	HashInsertText(svr,"stats","map",
		HashSelectText(svr,"items","mapname"));
	HashInsertInt(svr,"stats","numplayers",
		HSI(svr,"info","num_players") );
	HashInsertText(svr,"stats","maxplayers",
		HashSelectText(svr,"items","maxclients") );

	return 1; // update displayed info
}

int QUAKE2_IsFull(server_t* svr)
{
	if (HSI(svr,"info","num_players") && atoi( GetValue(svr, "maxclients") ) && 
			HSI(svr,"info","num_players") == atoi( GetValue(svr, "maxclients") ) )
		return 1;
	return 0;
}

char* QUAKE2_Cmdline(server_t* svr)
{
	static char string[256];
	strcpy(string,"");

	if (HashExistsKey(svr,"items","game")  != (HASH_PARAM)-1 && strcmp(HST(svr,"items","game"),"") != 0)
		sprintf(string+strlen(string),"+set game %s ",HashSelectText(svr,"items","game") );

	// set password if there's one
	if (strcmp(HashSelect(HASH_GLOBAL,"Status","Password",HASH_STRING),"") != 0)
			sprintf(string+strlen(string)," +set password \"%s\"",HashSelect(HASH_GLOBAL,"Status","Password",HASH_STRING));

	// and connect
	sprintf (string+strlen(string),"+connect %s", svr->ipaddr);
	return string;
}

int __stdcall Init(shared_t* info)
{
	types_t* t;

	GetServerAddr = info->GetServerAddr;
	GetServerPort = info->GetServerPort;

	HashInit(info->hashes,info->numHashes);
	s = HashSelectInt(HASH_GLOBAL,"Global","Socket");
	clan_id = HashSelectText(HASH_GLOBAL,"Settings","ClanId");
	
	t = info->types[0];
	//t->type = TYPE_QUAKE2;
	t->protocol = 1;
	t->port = 27910;
	strcpy(t->name, "Quake 2");
	strcpy(t->path, "C:\\Quake2\\Quake2.exe");
	strcpy(t->qstat_type, "Q2S");
	strcpy(t->gamespy_type, "quake2prot");
	strcpy(t->mod_var,"gamename");
	t->receive = QUAKE2_Receive;
	t->print_info = QUAKE2_PrintInfo;
	t->send = QUAKE2_Send;
	t->hostname = QUAKE2_Hostname;
	t->isfull = QUAKE2_IsFull;

/*	

	(void*)t->iImage = LoadBitmap(
		HashSelectPtr(HASH_GLOBAL,"Status","hInstance"),
		MAKEINTRESOURCE(IDB_GAMEICON)); */

	t = info->types[1];
	//t->type = TYPE_SOF;
	t->port = 28910;
	strcpy(t->name, "Soldier Of Fortune");
	strcpy(t->path, "C:\\SOF Demo");
	strcpy(t->qstat_type, "SFS");
	strcpy(t->gamespy_type, "soldieroffortune");
	
	t->receive = QUAKE2_Receive;
	t->print_info = QUAKE2_PrintInfo;
	t->send = QUAKE2_Send;
	t->hostname = QUAKE2_Hostname;
	t->isfull = QUAKE2_IsFull;

	(void*)t->iImage = LoadBitmap(
		HashSelectPtr(HASH_GLOBAL,"Status","hInstance"),
		MAKEINTRESOURCE(IDB_GAMEICON2));

	t = info->types[2];
	//t->type = 9;
	t->port = 31510;
	strcpy(t->name, "Kingpin: Life Of Crime");
	strcpy(t->path, "C:\\kingpin\\kingpin.exe");
	strcpy(t->qstat_type, "KPS");
	strcpy(t->gamespy_type, "kingpin");
	
	t->receive = QUAKE2_Receive;
	t->print_info = QUAKE2_PrintInfo;
	t->send = QUAKE2_Send;
	t->hostname = QUAKE2_Hostname;

	t->isfull = QUAKE2_IsFull;

	(void*)t->iImage = LoadBitmap(
		HashSelectPtr(HASH_GLOBAL,"Status","Hinstance"),
		MAKEINTRESOURCE(IDB_GAMEICON3)
	);

	t = info->types[3];
	//t->type = 2;
	t->port = 22450;
	strcpy(t->name, "Sin");
	strcpy(t->path, "C:\\Sin\\Sin.exe");
	strcpy(t->qstat_type, "SNS");
	strcpy(t->gamespy_type, "sin");
	
	t->receive = QUAKE2_Receive;
	t->print_info = QUAKE2_PrintInfo;
	t->send = QUAKE2_Send;
	t->hostname = QUAKE2_Hostname;
	t->isfull = QUAKE2_IsFull;

	(void*)t->iImage = LoadBitmap(
		HashSelectPtr(HASH_GLOBAL,"Status","hInstance"),
		MAKEINTRESOURCE(IDB_GAMEICON4)
	);

	t = info->types[4];
	//t->type = TYPE_DAIK;
	t->port = 27992;
	strcpy(t->name, "Daikatana");
	strcpy(t->path, "C:\\Daikatana\\Daikatana.exe");
	strcpy(t->qstat_type, "DKS");
	strcpy(t->gamespy_type, "daikatana");
	
	t->receive = QUAKE2_Receive;
	t->print_info = QUAKE2_PrintInfo;
	t->send = QUAKE2_Send;
	//t->cmdline = QUAKE2_Cmdline;
	t->hostname = QUAKE2_Hostname;
	t->isfull = QUAKE2_IsFull;
	

	(void*)t->iImage = LoadBitmap(
		HashSelectPtr(HASH_GLOBAL,"Status","hInstance"),
		MAKEINTRESOURCE(IDB_GAMEICON5)
	);
		
		
	return 5;
}

