#include <stdio.h>
#include "resource.h"
#include "sin\resource.h"
#include "types.h"
#include "shared.h"
#include "players.h"
#include "quakex.h"

SOCKET s;
char* clan_id;
unsigned long (*GetServerAddr)(char* addr);
short int (*GetServerPort)(char* addr, short otherwise);

void SIN_Send(server_t* svr, short port) 
{
	SOCKADDR_IN src_addr;
	
	src_addr.sin_family = AF_INET;
	if (svr) {
		src_addr.sin_addr.s_addr = GetServerAddr(svr->ipaddr);
		src_addr.sin_port = htons(GetServerPort(svr->ipaddr, port));
	}
	else {
		src_addr.sin_addr.s_addr = INADDR_BROADCAST;
		src_addr.sin_port = htons(port);
	}
	
	sendto(s, "\xff\xff\xff\xffstatus", 11, 0, (SOCKADDR*)&src_addr, sizeof(SOCKADDR_IN));
	
	return;
}

void SIN_PrintInfo(server_t* svr, HWND mainwnd, int ignore_changes)
{
	player_t temp_p;
	HWND temp, hwnd;
	int i, j;
	char buf[128];
	int sin = 0;

	hwnd = mainwnd;

	SetDlgItemText(hwnd, IDT_HOSTNAME, GetValue(svr, "hostname") );
	if (svr->timeout >= 5) {
		SetDlgItemText(hwnd, IDT_HOSTNAME, "UNREACHABLE");
		HashDeleteAll(svr,"items"); // wipe all current settings so we get blanks
		svr->num_players = 0;
	} else if (HashCount(svr,"items") == 0) {
		svr->num_players = 0;
		SetDlgItemText(hwnd, IDT_HOSTNAME, "CONNECTING...");
	}

	SetDlgItemText(hwnd, IDT_MAPNAME, GetValue(svr, "mapname") );
	if (!strcmp( GetValue(svr, "mapname"), "") || GetValue(svr, "mapname") == NULL)
	{
		SetDlgItemText(hwnd, IDT_MAPNAME, GetValue(svr, "map") );
	}
	
	SetDlgItemText(hwnd, IDT_FRAGLIMIT, GetValue(svr, "fraglimit") );
	SetDlgItemText(hwnd, IDT_TIMELIMIT, GetValue(svr, "timelimit") );
	SetDlgItemText(hwnd, IDC_CAPTURELIMIT, "Capturelimit:");
	SetDlgItemText(hwnd, IDT_CAPTURELIMIT, GetValue(svr, "capturelimit") );

	if (strcmp(GetValue(svr, "protocol"),""))
	{
		sscanf(GetValue(svr, "protocol"), "%d", &i);
		switch (i) {
			case 3:	 SetDlgItemText(hwnd, IDT_VERSION, "SIN  (v1.0)"); break;
			case 4:	 SetDlgItemText(hwnd, IDT_VERSION, "SIN  (v1.01)"); break;
			default: SetDlgItemText(hwnd, IDT_VERSION, "SIN");
		}
	} else 
		if (strcmp(GetValue(svr, "map"),""))
			SetDlgItemText(hwnd, IDT_VERSION, "SIN Demo  (v0.9)");
	else SetDlgItemText(hwnd, IDT_VERSION, "");

	if (strcmp(GetValue(svr, "maxclients"), ""))
	{
		sprintf(buf, "%d / %s", svr->num_players, GetValue(svr, "maxclients") );
		SetDlgItemText(hwnd, IDT_PLAYERS, buf);
	}
	else SetDlgItemText(hwnd, IDT_PLAYERS, "");

	SetDlgItemText(hwnd, IDC_DMFLAGS, "DM Flags:" );
	SetDlgItemText(hwnd, IDT_DMFLAGS, GetValue(svr, "dmflags") );

	SetDlgItemText(hwnd,IDT_MISC,"");

	temp = GetDlgItem(hwnd, IDL_PLAYERS);

	SendMessage(temp, LB_RESETCONTENT, 0, 0);
	
	i = 0;
	while (i < svr->num_players - 1) 
	{
		if (svr->players[i].frags < svr->players[i+1].frags)
		{
			temp_p = svr->players[i];
			svr->players[i] = svr->players[i+1];
			svr->players[i+1] = temp_p;
			i = 0;
			continue;
		}
		i++;
	}
	
	SendMessage(temp, LB_ADDSTRING, 0, (LPARAM)"\tFrags\tPing\tName");
		
	for (i=0;i<svr->num_players;i++)
	{
		strcpy(buf,"");

		if (strcmp(clan_id, ""))
		{
			for (j=0;j<(int)(strlen(svr->players[i].name) - strlen(clan_id));j++)
				if (!memcmp(svr->players[i].name+j, clan_id, strlen(clan_id)))
					strcat(buf,"*");
		}
		
		strcat(buf, "\t");

		if (!ignore_changes && svr->players[i].change)
			sprintf(buf + strlen(buf), "%d (%c%d)\t%d\t%s", svr->players[i].frags, svr->players[i].change > 0 ? '+' : ' ', svr->players[i].change,svr->players[i].ping,svr->players[i].name);
		else sprintf(buf + strlen(buf), "%d\t%d\t%s", svr->players[i].frags, svr->players[i].ping,svr->players[i].name);
		SendMessage(temp, LB_ADDSTRING, 0, (LPARAM)buf);
	}
	return;
}

char* SIN_Hostname(server_t* svr)
{
	return GetValue(svr, "hostname");
}

int SIN_Receive(server_t* svr, char* buf, int len)
{
	buf[len] = 0;
	DecodeString(svr, buf, len);
	return 1; // update displayed info
}

int QUAKE2_IsFull(server_t* svr)
{
	if (svr->num_players && atoi( GetValue(svr, "sv_maxclients") ) &&
			svr->num_players == atoi( GetValue(svr, "sv_maxclients") ) )
		return 1;
	return 0;
}
int __stdcall Init(shared_t* info)
{
	types_t* t;

	GetServerAddr = info->GetServerAddr;
	GetServerPort = info->GetServerPort;

	HashInit(info->hashes,info->numHashes);
	s = HashSelectInt(HASH_GLOBAL,"Global","Socket");
	clan_id = HashSelectText(HASH_GLOBAL,"Settings","ClanId");
	
	t = info->types[0];
	t->type = 2;
	t->port = 22450;
	strcpy(t->name, "Sin");
	strcpy(t->path, "C:\\Sin\\Sin.exe");
	t->receive = SIN_Receive;
	t->print_info = SIN_PrintInfo;
	t->send = SIN_Send;
	t->hostname = SIN_Hostname;

	HashInsertPtr(t,"funcs","isfull",QUAKE2_IsFull);

	(void*)t->iImage = LoadBitmap(
		HashSelectPtr(HASH_GLOBAL,"Status","hInstance"),
		MAKEINTRESOURCE(IDB_GAMEICON)
	);

	return 1;
}

