/*
******************************************************************
Server Query; a free games server monitoring tool
Copyright (C) 1998-2002  Nathan O'Sullivan

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
******************************************************************
*/
#include <stdio.h>
#include "resource.h"
#include "tribes\resource.h"
#include "types.h"
#include "shared.h"
#include "players.h"
#include "servinfo.h"


#define SVR_NUMPLAYERS (HSI(svr,"info","num_players"))
#define SVR_FRAGS(x) (HSI(svr,"frags",(HASH_PARAM)x))
#define SVR_NAME(x) (HST(svr,"name",(HASH_PARAM)x))
#define SVR_PING(x) (HSI(svr,"ping",(HASH_PARAM)x))
#define SVR_CHANGE(x) (HSI(svr,"change",(HASH_PARAM)x))
#define SVR_TEAM(x) (HSI(svr,"team",(HASH_PARAM)x))
#define SVR_ISTEAM(x) (HSI(svr,"isteam",(HASH_PARAM)x))

SOCKET s;
char* clan_id;
unsigned long (*GetServerAddr)(char* addr);
short int (*GetServerPort)(char* addr, short otherwise);

#define IS_TEAM other[0]
#define TEAM other[1]

void TRIBES_Send(SOCKET s, server_t* svr, short port) 
{
	SOCKADDR_IN src_addr;
	char* str;

	src_addr.sin_family = AF_INET;
	if (svr) {
		src_addr.sin_addr.s_addr = GetServerAddr(svr->ipaddr);
		src_addr.sin_port = htons(GetServerPort(svr->ipaddr, port));
	}
	else {
		src_addr.sin_addr.s_addr = INADDR_BROADCAST;
		src_addr.sin_port = htons(port);
	}
	
	str = "\x62\0\0";
	sendto(s, str, 3, 0, (SOCKADDR*)&src_addr, sizeof(SOCKADDR_IN));
	return;
}

void TRIBES_PrintInfo(server_t* svr, HWND mainwnd, int ignore_changes)
{
//	player_t temp_p;
	HWND temp, hwnd;
	int i, j, numteams, team;
	char buf[128];
	int sin = 0;
	int order[128];

	hwnd = mainwnd;

	ServinfoInit(svr);

	strcpy(buf, GetValue(svr, "hostname") );
	if (svr->timeout >= 5) {
		strcpy(buf, "UNREACHABLE");
		HashDeleteAll(svr,"items"); // wipe all current settings so we get blanks
		HII(svr,"info","num_players",0);
	} else if (HashCount(svr,"items") == 0) {
		HII(svr,"info","num_players",0);
		strcpy(buf, "CONNECTING...");
	}
	SetDlgItemText(hwnd, IDT_HOSTNAME, buf );
	ServinfoAdd("Hostname",buf);
	ServinfoAdd("Address", svr->ipaddr);

	if (GetValue(svr,"version")[0])
		sprintf(buf,"Tribes  (v%s)",GetValue(svr,"version"));
	else sprintf(buf,"Tribes");
	SetDlgItemText(hwnd, IDT_VERSION, buf);
	ServinfoAdd("Game",buf);

	strcpy(buf, GetValue(svr, "mapname") );
	SetDlgItemText(hwnd, IDT_MAPNAME, buf );
	ServinfoAdd("Map",buf);

	if (strcmp(GetValue(svr, "maxplayers"), ""))
	{
		sprintf(buf, "%s / %s", GetValue(svr, "numplayers"), GetValue(svr, "maxplayers") );
	}
	else if (strcmp(GetValue(svr, "numplayers"), ""))
	{
		sprintf(buf, "%s", GetValue(svr, "numplayers"));
	} else
		strcpy(buf,"");
	SetDlgItemText(hwnd, IDT_PLAYERS, buf);
	ServinfoAdd("Players",buf);

	ServinfoAdd("Ping","");

  if (HSIG("Settings","ExtraServInfo")) {
	strcpy(buf, GetValue(svr, "mp_fraglimit") );
	SetDlgItemText(hwnd, IDT_FRAGLIMIT, buf );
	ServinfoAdd("Fraglimit",buf);

	strcpy(buf, GetValue(svr, "mp_timelimit") );
	SetDlgItemText(hwnd, IDT_TIMELIMIT, buf );
	ServinfoAdd("Timelimit",buf);

	SetDlgItemText(hwnd, IDC_CAPTURELIMIT, "Game style:");
	strcpy(buf, GetValue(svr, "gametype") );
	SetDlgItemText(hwnd, IDT_CAPTURELIMIT, buf );
	ServinfoAdd("Game style",buf);

/*	SetDlgItemText(hwnd, IDC_DMFLAGS, "DM Flags:" );
	strcpy(buf, GetValue(svr, "dmflags") );
	SetDlgItemText(hwnd, IDT_DMFLAGS, buf );
	ServinfoAdd("DM Flags",buf);*/

	SetDlgItemText(hwnd,IDT_MISC,"");
  }

    PlayerinfoInit(svr);
	temp = GetDlgItem(hwnd, IDL_PLAYERS);

	SendMessage(temp, LB_RESETCONTENT, 0, 0);
	
	//** SORT THE TEAMS AND PLAYERS **//
	numteams = 0;
	sscanf(GetValue(svr,"numteams"),"%d",&numteams);


	for (i=0;i<SVR_NUMPLAYERS;i++)
		order[i] = i;

	i = 0;
	while (i < numteams - 1) 
	{
		if (SVR_FRAGS(order[i]) < SVR_FRAGS(order[i+1]))
		{
			j = order[i];
			order[i] = order[i+1];
			order[i+1] = j;
//			temp_p = svr->players[i];
			i = 0;
			continue;
		}
		i++;
	}
	i = numteams;

	while (i < SVR_NUMPLAYERS - 1) 
	{
		if (SVR_FRAGS(order[i]) < SVR_FRAGS(order[i+1]))
		{
			j = order[i];
			order[i] = order[i+1];
			order[i+1] = j;
			i = numteams;
			continue;
		}
		i++;
	}
	
		
	for (team=0;team<numteams;team++)
	{
		sprintf(buf,"Team \"%s\" (%d points)",SVR_NAME(order[team]),SVR_FRAGS(order[team]));
		SendMessage(temp, LB_ADDSTRING, 0, (LPARAM)buf);
		strcpy(buf,SVR_NAME(order[team]));
		PlayerinfoAdd(buf,-3000 - SVR_FRAGS(order[team]),0,-2000);
		
		for (i=numteams;i<SVR_NUMPLAYERS;i++)
		{
			if (SVR_TEAM(order[i]) == team + 1)
			{	
				strcpy(buf,"");

				if (strcmp(clan_id,""))
				{
					for (j=0;j<(int)(strlen(SVR_NAME(order[i])) - strlen(clan_id));j++)
						if (!memcmp(SVR_NAME(order[i])+j, clan_id, strlen(clan_id)))
							strcat(buf,"*");
				}
		
				strcat(buf, "\t");

				if (!ignore_changes && SVR_CHANGE(order[i]))
					sprintf(buf + strlen(buf), "%d (%c%d)\t%d\t%s", 
						SVR_FRAGS(order[i]),
						SVR_CHANGE(order[i]) > 0 ? '+' : ' ',
						SVR_CHANGE(order[i]),
						SVR_PING(order[i]),
						SVR_NAME(order[i])
					);
				else 
					sprintf(buf + strlen(buf), "%d\t%d\t%s", 
						SVR_FRAGS(order[i]),
						SVR_PING(order[i]),
						SVR_NAME(order[i])
					);
				
				SendMessage(temp, LB_ADDSTRING, 0, (LPARAM)buf);
				PlayerinfoAdd(SVR_NAME(order[i]),SVR_FRAGS(order[i]),SVR_CHANGE(order[i]),SVR_PING(order[i]));
			}
		}
		if (SVR_NUMPLAYERS <= 20) {
			SendMessage(temp, LB_ADDSTRING, 0, (LPARAM)"");
			PlayerinfoAdd("",-2000,0,-2000);
		}
	}
	
	j=0;
	for (i=numteams;i<SVR_NUMPLAYERS;i++)
		if (SVR_TEAM(order[i]) == 0)
			j++;

	if (j) {
		SendMessage(temp,LB_ADDSTRING, 0, (LPARAM)"Observers");
		PlayerinfoAdd("Observers",-2000,0,-2000);
	}

	for (i=numteams;i<SVR_NUMPLAYERS;i++)
		if (SVR_TEAM(order[i]) == 0)
		{
			strcpy(buf,"");

			if (strcmp(clan_id,""))
			{
				for (j=0;j<(int)(strlen(SVR_NAME(order[i])) - strlen(clan_id));j++)
					if (!memcmp(SVR_NAME(order[i])+j, clan_id, strlen(clan_id)))
						strcat(buf,"*");
			}

			strcat(buf, "\t");
			sprintf(buf+strlen(buf), "%d\t%s", SVR_PING(order[i]),SVR_NAME(order[i]));
				SendMessage(temp, LB_ADDSTRING, 0, (LPARAM)buf);
			PlayerinfoAdd(SVR_NAME(order[i]),-2000,SVR_CHANGE(order[i]),SVR_PING(order[i]));
		}

	
		
	return;
}

char* TRIBES_Hostname(server_t* svr)
{
	return GetValue(svr, "hostname");
}

int GetString(char* ptr, char* buf, int max)
{
	int i = 0;
	int length = *((unsigned char*)ptr);
	ptr++;

	while (i < max && i < length)
	{
		*(buf++) = *(ptr++);
		i++;
	}
	*buf = 0;
	return i + 1;
}

void ExpandAmpersand(char* str)
{
	char buf[256];
	int i;

	strcpy (buf, str);
	for (i=0;i<(int)strlen(buf);i++)
	{
		*str = buf[i];
		/*
		if (buf[i] == '&')
		{
			memcpy(str," && ", 4);
			str+=4;
			*str = ' ';
		}
		else
		*/
			str++;
		
	}
}

int TRIBES_Receive(server_t* svr, char* buf, int len)
{
	char* ptr;
	char buffer[257];
	int i, j, numteams;
	char value[128];

	ptr = buf;
	if (memcmp(ptr, "\x63\x0\x0\x62",4))
		return 0;
	
	HashDeleteAll(svr,"items");

	ptr+=4;
	ptr += GetString(ptr,buffer,128); // "Tribes" <-- gamename ?
	
	ptr += GetString(ptr,value,256); 
	HashInsertText(svr,"items","version", value);

	ptr += GetString(ptr,value,256); 
	HashInsertText(svr,"items","hostname", value);

	ptr += GetString(ptr,buffer,128); // unknown str ? is it actually a string [hex was 2 bytes - 0x0101]
	
	sprintf(value, "%d", (char)ptr[0]);
	HashInsertText(svr,"items","numplayers", value);

	sprintf(value, "%d", (char)ptr[1]);
	HashInsertText(svr,"items","maxplayers", value);

	ptr+=2;

	ptr+=2; // unknown

	ptr += GetString(ptr,buffer,128); // unknown str
	
	ptr += GetString(ptr,value,256); 
	ExpandAmpersand(value);
	HashInsertText(svr,"items","gametype", value);
	
	ptr += GetString(ptr,value,256); 
	HashInsertText(svr,"items","mapname", value);
	
	ptr += GetString(ptr,buffer,256); // unnecessary server email + server info

	numteams = *ptr;
	ptr++;
	sprintf(value, "%d", numteams);
	HashInsertText(svr,"items","numteams", value);
	
	ptr += GetString(ptr,value,256); 
	HashInsertText(svr,"items","teamheader", value);

	ptr += GetString(ptr,value,256); 
	HashInsertText(svr,"items","playerheader", value);

	HII(svr,"info","num_players", 0);

	for (i=0;i<numteams;i++)
	{
		ptr+= GetString(ptr,value, 256);
		HIT(svr,"name",(HASH_PARAM)SVR_NUMPLAYERS,value);
		ptr += GetString(ptr,buffer,256); // strip the '%t'
		//ptr += GetString(ptr,buffer,256); // get the score
		sscanf(buffer+3, "%d", &j);
		HII(svr,"frags",(HASH_PARAM)SVR_NUMPLAYERS,j);
		HII(svr,"isteam",(HASH_PARAM)SVR_NUMPLAYERS,1);
		HII(svr,"info","num_players",SVR_NUMPLAYERS + 1);
	}

	while (ptr < buf + len)
	{
		HII(svr,"ping",(HASH_PARAM)SVR_NUMPLAYERS, *(unsigned char*)(ptr) * 4 );
		HII(svr,"team",(HASH_PARAM)SVR_NUMPLAYERS, *(ptr+2) + 1 );
		
		//svr->players[svr->num_players].ping = *(unsigned char*)(ptr) * 4;
		//svr->players[svr->num_players].TEAM = *(ptr+2) + 1;
		ptr += 3;
		ptr += GetString(ptr,buffer,256); 
		//strcpy(svr->players[svr->num_players].name,buffer);
		HIT(svr,"name",(HASH_PARAM)SVR_NUMPLAYERS,buffer);
		ptr += GetString(ptr,buffer,256); 
		sscanf(buffer+6,"%d",&j);
		HII(svr,"frags",(HASH_PARAM)SVR_NUMPLAYERS,j);
		HII(svr,"info","num_players",SVR_NUMPLAYERS + 1);
		//svr->num_players++;
	}
	CleanRecords(svr);

	// Update available information
	HashInsertText(svr,"stats","map",
		HashSelectText(svr,"items","mapname"));
	HashInsertText(svr,"stats","numplayers",
		HashSelectText(svr,"items","numplayers"));
	HashInsertText(svr,"stats","maxplayers",
		HashSelectText(svr,"items","maxplayers") );

	return 1; // update displayed info
}

char* TRIBES_Cmdline(server_t* svr)
{
	static char string[256];
	if (!strchr(svr->ipaddr,':'))
		sprintf(string, "+connect %s:28001", svr->ipaddr);
	else sprintf(string, "+connect %s", svr->ipaddr);
	return string;
}

int TRIBES_IsFull(server_t* svr)
{
	if (atoi( GetValue(svr,"numplayers") ) && atoi( GetValue(svr, "maxplayers") ) && 
			atoi ( GetValue(svr,"numplayers") ) == atoi( GetValue(svr, "maxplayers") ) )
		return 1;
	return 0;
}
int __stdcall Init(shared_t* info)
{
	types_t* t;

	GetServerAddr = info->GetServerAddr;
	GetServerPort = info->GetServerPort;

	HashInit(info->hashes,info->numHashes);
	s = HashSelectInt(HASH_GLOBAL,"Global","Socket");
	clan_id = HashSelectText(HASH_GLOBAL,"Settings","ClanId");
	

	t = info->types[0];
	//t->type = 5;
	t->port = 28001;
	strcpy(t->name, "Tribes");
	strcpy(t->path, "C:\\Tribes\\tribes.exe");
	strcpy(t->qstat_type, "TBS");
	strcpy(t->gamespy_type, "tribes");
	strcpy(t->mod_var,"game");
	
	t->receive = TRIBES_Receive;
	t->print_info = TRIBES_PrintInfo;
	t->send = TRIBES_Send;
	t->hostname = TRIBES_Hostname;
	
	t->isfull = TRIBES_IsFull;

	(void*)t->iImage = LoadBitmap(
		HashSelectPtr(HASH_GLOBAL,"Status","hInstance"),
		MAKEINTRESOURCE(IDB_GAMEICON)
	);

	return 1;
}

