/*
******************************************************************
Server Query; a free games server monitoring tool
Copyright (C) 1998-2002  Nathan O'Sullivan

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
******************************************************************
*/
#include <stdio.h>
#include "resource.h"
#include "tribes2\resource.h"
#include "types.h"
#include "shared.h"
#include "players.h"
#include "servinfo.h"


#define SVR_NUMPLAYERS (HSI(svr,"info","num_players"))
#define SVR_FRAGS(x) (HSI(svr,"frags",(HASH_PARAM)x))
#define SVR_NAME(x) (HST(svr,"name",(HASH_PARAM)x))
#define SVR_PING(x) (HSI(svr,"ping",(HASH_PARAM)x))
#define SVR_CHANGE(x) (HSI(svr,"change",(HASH_PARAM)x))
#define SVR_TEAM(x) (HSI(svr,"team",(HASH_PARAM)x))
#define SVR_ISTEAM(x) (HSI(svr,"isteam",(HASH_PARAM)x))

#define TRIBES2_STATUS_DEDICATED	(1<<0)
#define TRIBES2_STATUS_PASSWORD		(1<<1)
#define TRIBES2_STATUS_LINUX		(1<<2)
#define TRIBES2_STATUS_TOURNAMENT	(1<<3)
#define TRIBES2_STATUS_NOALIAS		(1<<4)
#define TRIBES2_STATUS_TOURNAMENT_VER3	(1<<6)
#define TRIBES2_STATUS_NOALIAS_VER3	(1<<7)


SOCKET s;
char* clan_id;
unsigned long (*GetServerAddr)(char* addr);
short int (*GetServerPort)(char* addr, short otherwise);
int (*GetAccTickCount)(void);

#define IS_TEAM other[0]
#define TEAM other[1]

void TRIBES2_Send(SOCKET s, server_t* svr, short port) 
{
	SOCKADDR_IN src_addr;
	char str1[] = { 0x0e, 0x02, 0x00, 0x00, 0x01, 0x00 };

	src_addr.sin_family = AF_INET;
	if (svr) {
		src_addr.sin_addr.s_addr = GetServerAddr(svr->ipaddr);
		src_addr.sin_port = htons(GetServerPort(svr->ipaddr, port));
	}
	else {
		src_addr.sin_addr.s_addr = INADDR_BROADCAST;
		src_addr.sin_port = htons(port);
	}
	
	sendto(s, str1, 6, 0, (SOCKADDR*)&src_addr, sizeof(SOCKADDR_IN));
	return;
}

void TRIBES2_SendInfoPacket(server_t* svr, short port) 
{
	SOCKADDR_IN src_addr;
	char str1[] = { 0x12, 0x02, 0x00, 0x00, 0x01, 0x00 };

	src_addr.sin_family = AF_INET;
	if (svr) {
		src_addr.sin_addr.s_addr = GetServerAddr(svr->ipaddr);
		src_addr.sin_port = htons(GetServerPort(svr->ipaddr, port));
		GetAccTickCount = HSPG("Funcs","GetAccTickCount");
		svr->last_req = GetAccTickCount();
	}
	else {
		src_addr.sin_addr.s_addr = INADDR_BROADCAST;
		src_addr.sin_port = htons(port);
	}
	
	
	sendto(s, str1, 6, 0, (SOCKADDR*)&src_addr, sizeof(SOCKADDR_IN));
	return;
}


void TRIBES2_PrintInfo(server_t* svr, HWND mainwnd, int ignore_changes)
{
//	player_t temp_p;
	HWND temp, hwnd;
	int i, j, numteams, team;
	char buf[128];
	int sin = 0;
	int order[128];

	hwnd = mainwnd;

	ServinfoInit(svr);

	strcpy(buf, GetValue(svr, "hostname") );
	
	if (svr->timeout >= 5) {
		strcpy(buf, "UNREACHABLE");
		HashDeleteAll(svr,"items"); // wipe all current settings so we get blanks
		HII(svr,"info","num_players",0);
	} 
	  else if (HashCount(svr,"items") == 0) {
		HII(svr,"info","num_players",0);
		strcpy(buf, "CONNECTING...");
	}
	SetDlgItemText(hwnd, IDT_HOSTNAME, buf );
	ServinfoAdd("Hostname",buf);
	ServinfoAdd("Address", svr->ipaddr);

	
	if (GetValue(svr,"version")[0])
		sprintf(buf,"Tribes2  (v%s)",GetValue(svr,"version"));
	else sprintf(buf,"Tribes2");
	SetDlgItemText(hwnd, IDT_VERSION, buf);
	ServinfoAdd("Game",buf);

	

	strcpy(buf, GetValue(svr, "mapname") );
	SetDlgItemText(hwnd, IDT_MAPNAME, buf );
	ServinfoAdd("Map",buf);

	if (strcmp(GetValue(svr, "maxplayers"), ""))
	{
		sprintf(buf, "%s / %s", GetValue(svr, "numplayers"), GetValue(svr, "maxplayers") );
	}
	else if (strcmp(GetValue(svr, "numplayers"), ""))
	{
		sprintf(buf, "%s", GetValue(svr, "numplayers"));
	} else
		strcpy(buf,"");
	
	//strcpy(buf,"Unknown");
	
	SetDlgItemText(hwnd, IDT_PLAYERS, buf); 
	ServinfoAdd("Players",buf);

	ServinfoAdd("Ping","");

	
  if (HSIG("Settings","ExtraServInfo")) {
	/*strcpy(buf, GetValue(svr, "mp_fraglimit") );
	SetDlgItemText(hwnd, IDT_FRAGLIMIT, buf );
	ServinfoAdd("Fraglimit",buf);

	strcpy(buf, GetValue(svr, "mp_timelimit") );
	SetDlgItemText(hwnd, IDT_TIMELIMIT, buf );
	ServinfoAdd("Timelimit",buf);*/

	SetDlgItemText(hwnd, IDC_CAPTURELIMIT, "Game style:");
	strcpy(buf, GetValue(svr, "mission") );
	SetDlgItemText(hwnd, IDT_CAPTURELIMIT, buf );
	ServinfoAdd("Game style",buf);

/*	SetDlgItemText(hwnd, IDC_DMFLAGS, "DM Flags:" );
	strcpy(buf, GetValue(svr, "dmflags") );
	SetDlgItemText(hwnd, IDT_DMFLAGS, buf );
	ServinfoAdd("DM Flags",buf);*/

	SetDlgItemText(hwnd,IDT_MISC,"");
  }

    PlayerinfoInit(svr);
	temp = GetDlgItem(hwnd, IDL_PLAYERS);

	SendMessage(temp, LB_RESETCONTENT, 0, 0);
	
	//** SORT THE TEAMS AND PLAYERS **//
	numteams = 0;
	sscanf(GetValue(svr,"numteams"),"%d",&numteams);


	for (i=0;i<SVR_NUMPLAYERS;i++)
		order[i] = i;

	i = 0;
	while (i < numteams - 1) 
	{
		if (SVR_FRAGS(order[i]) < SVR_FRAGS(order[i+1]))
		{
			j = order[i];
			order[i] = order[i+1];
			order[i+1] = j;
//			temp_p = svr->players[i];
			i = 0;
			continue;
		}
		i++;
	}
	i = numteams;

	while (i < SVR_NUMPLAYERS - 1) 
	{
		if (SVR_FRAGS(order[i]) < SVR_FRAGS(order[i+1]))
		{
			j = order[i];
			order[i] = order[i+1];
			order[i+1] = j;
			i = numteams;
			continue;
		}
		i++;
	}
	
		
	for (team=0;team<numteams;team++)
	{
		sprintf(buf,"Team \"%s\" (%d points)",SVR_NAME(order[team]),SVR_FRAGS(order[team]));
		SendMessage(temp, LB_ADDSTRING, 0, (LPARAM)buf);
		strcpy(buf,SVR_NAME(order[team]));
		PlayerinfoAdd(buf,-3000 - SVR_FRAGS(order[team]),0,-2000);
		
		for (i=numteams;i<SVR_NUMPLAYERS;i++)
		{
			if (SVR_TEAM(order[i]) == team + 1)
			{	
				strcpy(buf,"");

				if (strcmp(clan_id,""))
				{
					for (j=0;j<(int)(strlen(SVR_NAME(order[i])) - strlen(clan_id));j++)
						if (!memcmp(SVR_NAME(order[i])+j, clan_id, strlen(clan_id)))
							strcat(buf,"*");
				}
		
				strcat(buf, "\t");

				if (!ignore_changes && SVR_CHANGE(order[i]))
					sprintf(buf + strlen(buf), "%d (%c%d)\t%d\t%s", 
						SVR_FRAGS(order[i]),
						SVR_CHANGE(order[i]) > 0 ? '+' : ' ',
						SVR_CHANGE(order[i]),
						SVR_PING(order[i]),
						SVR_NAME(order[i])
					);
				else 
					sprintf(buf + strlen(buf), "%d\t%d\t%s", 
						SVR_FRAGS(order[i]),
						SVR_PING(order[i]),
						SVR_NAME(order[i])
					);
				
				SendMessage(temp, LB_ADDSTRING, 0, (LPARAM)buf);
				PlayerinfoAdd(SVR_NAME(order[i]),SVR_FRAGS(order[i]),SVR_CHANGE(order[i]),-1000); //SVR_PING(order[i]));
			}
		}
		if (SVR_NUMPLAYERS <= 20) {
			SendMessage(temp, LB_ADDSTRING, 0, (LPARAM)"");
			PlayerinfoAdd("",-2000,0,-2000);
		}
	}
	
	j=0;
	for (i=numteams;i<SVR_NUMPLAYERS;i++)
		if (SVR_TEAM(order[i]) == 0)
			j++;

	if (j) {
		SendMessage(temp,LB_ADDSTRING, 0, (LPARAM)"Observers");
		PlayerinfoAdd("Observers",-2000,0,-2000);
	}

	for (i=numteams;i<SVR_NUMPLAYERS;i++)
		if (SVR_TEAM(order[i]) == 0)
		{
			strcpy(buf,"");

			if (strcmp(clan_id,""))
			{
				for (j=0;j<(int)(strlen(SVR_NAME(order[i])) - strlen(clan_id));j++)
					if (!memcmp(SVR_NAME(order[i])+j, clan_id, strlen(clan_id)))
						strcat(buf,"*");
			}

			strcat(buf, "\t");
			sprintf(buf+strlen(buf), "%d\t%s", SVR_PING(order[i]),SVR_NAME(order[i]));
				SendMessage(temp, LB_ADDSTRING, 0, (LPARAM)buf);
			PlayerinfoAdd(SVR_NAME(order[i]),-2000,SVR_CHANGE(order[i]),SVR_PING(order[i]));
		}

	
		
	return;
}

char* TRIBES2_Hostname(server_t* svr)
{
	return GetValue(svr, "hostname");
	//return "Tribes2 Server";
}

int GetNLString(char* ptr, char* buf, int max)
{
	int i;

	buf[0] = 0;

	
	for (i=0;i<max;i++)
		if (ptr[i] == 10 || !ptr[i])
			break;
	memcpy(buf,ptr,i);
	buf[i] = 0;
	return i + (ptr[i] == 10 ? 1 : 0);
}

int GetString(char* ptr, char* buf, int max)
{
	int i = 0;
	int length = *((unsigned char*)ptr);
	ptr++;

	while (i < max && i < length)
	{
		*(buf++) = *(ptr++);
		i++;
	}
	*buf = 0;
	return i + 1;
}

int GetLong(char* ptr, long* val)
{
	memcpy(val,ptr,4);
	return 4;
}

int GetShort(char* ptr, short* val)
{
	memcpy(val,ptr,2);
	return 2;
}

void ExpandAmpersand(char* str)
{
	char buf[256];
	int i;

	strcpy (buf, str);
	for (i=0;i<(int)strlen(buf);i++)
	{
		*str = buf[i];
		/*
		if (buf[i] == '&')
		{
			memcpy(str," && ", 4);
			str+=4;
			*str = ' ';
		}
		else
		*/
			str++;
		
	}
}

#define BUFSIZE	256

int TRIBES2_ReceivePingPacket(server_t* svr, char* inputbuf, int len)
{
	char* ptr;
	char buf[BUFSIZE+1];
//	int i;
	long value;

	HashDeleteAll(svr,"items");

	ptr = inputbuf;
	
	ptr += 6; // skip header info
	ptr += GetString(ptr,buf,BUFSIZE); // query protocol version
	HashInsertText(svr,"items","queryversion", buf);

	if (buf[3] >= '5')	//version 5 or greater
	{
		ptr += GetLong(ptr,&value);
		HashInsertInt(svr,"items","net_protocol",value);
	}

	ptr += GetLong(ptr,&value);
	HashInsertInt(svr,"items","minimum_net_protocol",value);

	ptr += GetLong(ptr,&value);
	HashInsertInt(svr,"items","build_version",value);

	ptr += GetString(ptr,buf,BUFSIZE);
	HashInsertText(svr,"items","hostname",buf);

	TRIBES2_SendInfoPacket(svr, 28000);

	return 0;
}

#define OUTPOS(x) if (ptr-inputbuf > len) { sprintf(output,"Buffer overflow at %s (at %d, len %d)",x,ptr-inputbuf,len); WRITE_DEBUG(output); TRIBES2_UpdateStats(svr); return 1; } // sprintf(output,"%s: At %d, len %d",x,ptr-inputbuf,len); WRITE_DEBUG(output)

void TRIBES2_UpdateStats(server_t* svr)
{
	CleanRecords(svr);

	// Update available information
	HashInsertText(svr,"stats","map",
		HashSelectText(svr,"items","mapname"));
	HashInsertText(svr,"stats","numplayers",
		HashSelectText(svr,"items","numplayers"));
	HashInsertText(svr,"stats","maxplayers",
		HashSelectText(svr,"items","maxplayers") );
	
	return;
}

int TRIBES2_ReceiveInfoPacket(server_t* svr, char* inputbuf, int len)
{
	char output[128];
	char* ptr;
	char buf[BUFSIZE+1];
	int i, j, k, limit;
	long value; 
	short shortval;
	short data;
	int numteams;
	int numplayers;
	char name[BUFSIZE+1];
	char teamname[BUFSIZE+1];
	int teamid;
	
	
	unsigned char status;

	
	ptr = inputbuf;
	
	ptr += 6; // skip header info
	OUTPOS("header");

	ptr += GetString(ptr,buf,BUFSIZE);
	HashInsertText(svr,"items","game",buf);
	OUTPOS("game");

	ptr += GetString(ptr,buf,BUFSIZE);
	HashInsertText(svr,"items","mission",buf);
	OUTPOS("mission");

	ptr += GetString(ptr,buf,BUFSIZE);
	HashInsertText(svr,"items","mapname",buf);
	OUTPOS("map");

	// this byte contains information on various server settings
	status = *(unsigned char*)ptr;
	if (status & TRIBES2_STATUS_DEDICATED)
		HashInsertInt(svr,"items","dedicated",1);
	if (status & TRIBES2_STATUS_PASSWORD)
		HashInsertInt(svr,"items","need_password",1);
	if (status & TRIBES2_STATUS_LINUX)
		HashInsertInt(svr,"items","linux",1);

	if (strcmp(HST(svr,"items","queryversion"),"VER3") == 0)
	{
		if (status & TRIBES2_STATUS_TOURNAMENT_VER3)
			HashInsertInt(svr,"items","tournament",1);
		if (status & TRIBES2_STATUS_NOALIAS_VER3)
			HashInsertInt(svr,"items","noalias",1);
	}
	else
	{
		if (status & TRIBES2_STATUS_TOURNAMENT)
			HashInsertInt(svr,"items","tournament",1);
		if (status & TRIBES2_STATUS_NOALIAS)
			HashInsertInt(svr,"items","noalias",1);
	}
	ptr ++;
	OUTPOS("status");

	HashInsertInt(svr,"items","numplayers",*(unsigned char*)ptr); ptr ++;
	HashInsertInt(svr,"items","maxplayers",*(unsigned char*)ptr); ptr ++;
	HashInsertInt(svr,"items","numbots",*(unsigned char*)ptr); ptr ++;
	OUTPOS("players/bots");
	
	ptr += GetShort(ptr,&shortval); 
	HashInsertInt(svr,"items","cpuspeed",shortval);
	OUTPOS("cpuspeed");

	ptr += GetString(ptr,buf,BUFSIZE);
	HashInsertText(svr,"items","serverinfo",buf);
	OUTPOS("serverinfo");

	ptr += GetShort(ptr,&data);	// Amount of data still to be read
	HII(svr,"info","num_players", 0);
	OUTPOS("start players");

	
	// Read # of teams
	ptr += GetNLString(ptr,buf,BUFSIZE);
	numteams = 0;
	value = sscanf(buf,"%d",&numteams);
	HashInsertInt(svr,"items","numteams",numteams);

	for (i=0;i<numteams;i++)
	{
		// Read team name and score
		ptr += GetNLString(ptr,buf,BUFSIZE);

		sscanf(buf,"%[^\t]\t%d",name,&value);

		HIT(svr,"name",(HASH_PARAM)SVR_NUMPLAYERS,name);
		HII(svr,"frags",(HASH_PARAM)SVR_NUMPLAYERS,value);
		HII(svr,"isteam",(HASH_PARAM)SVR_NUMPLAYERS,1);
		HII(svr,"info","num_players",SVR_NUMPLAYERS + 1);
	}
	
	if (numteams == 0)
	{
		HIT(svr,"name",(HASH_PARAM)SVR_NUMPLAYERS,"Storm");
		HIT(svr,"frags",(HASH_PARAM)SVR_NUMPLAYERS,"?");
		HII(svr,"isteam",(HASH_PARAM)SVR_NUMPLAYERS,1);
		HII(svr,"info","num_players",SVR_NUMPLAYERS + 1);
		
		HIT(svr,"name",(HASH_PARAM)SVR_NUMPLAYERS,"Inferno");
		HIT(svr,"frags",(HASH_PARAM)SVR_NUMPLAYERS,"?");
		HII(svr,"isteam",(HASH_PARAM)SVR_NUMPLAYERS,1);
		HII(svr,"info","num_players",SVR_NUMPLAYERS + 1);

		numteams = 2;
		HashInsertInt(svr,"items","numteams",2);
	}
	OUTPOS("read teams");

	// Read # of players
	ptr += GetNLString(ptr,buf,BUFSIZE);
	numplayers = 0;
	sscanf(buf,"%d",&numplayers);

	WRITE_DEBUG("Number of players:");
	WRITE_DEBUG(buf);
	
	for (i=0;i<numplayers;i++)
	{
		ptr += GetNLString(ptr,buf,BUFSIZE);
		sscanf(buf,"%[^\t]\t%[^\t]\t%d",name,teamname,&value);

		// first two and last character should be ignored
		//name[strlen(name)-1] = 0;


		for (j=0;j<(int)strlen(name);j++)
		{
			if (name[j] < 32 && name[j] != 0)
			{
				for (k=j,limit=strlen(name);k<limit;k++)
					name[k] = name[k+1];
				j--;
			}
		}


		HIT(svr,"name",(HASH_PARAM)SVR_NUMPLAYERS,name);
		HIT(svr,"ping",(HASH_PARAM)SVR_NUMPLAYERS, "?" );
		HII(svr,"frags",(HASH_PARAM)SVR_NUMPLAYERS,value);

		teamid = 0;
		for (j=0;j<numteams;j++)
		{
			if (strcmp(HST(svr,"name",(HASH_PARAM)j),teamname) == 0)
				teamid = j + 1;
		}	
		HII(svr,"team",(HASH_PARAM)SVR_NUMPLAYERS,teamid);
		HII(svr,"info","num_players",SVR_NUMPLAYERS + 1);
		OUTPOS("read player");
	}

	WRITE_DEBUG("Done");

	TRIBES2_UpdateStats(svr);

			

	return 1;
}

int TRIBES2_Receive(server_t* svr, char* buf, int len)
{
	char* ptr;
	/*char buffer[257];
	int i, j, numteams;
	char value[128];*/
	
	ptr = buf;
	ptr[len] = 0;

	if (*ptr == 16)
		return TRIBES2_ReceivePingPacket(svr,buf,len);
	else if (*ptr == 20)
		return TRIBES2_ReceiveInfoPacket(svr,buf,len);
	
	return 0;
	
}

char* TRIBES2_Cmdline(server_t* svr)
{
	static char string[256];
	strcpy(string, "");

	if (svr->lan)
		strcat(string," -nologin");

	if (!strchr(svr->ipaddr,':'))
		sprintf(string+strlen(string), " -connect IP:%s:28000", svr->ipaddr);
	else 
		sprintf(string+strlen(string), " -connect IP:%s", svr->ipaddr);


	return string;
}

int TRIBES2_IsFull(server_t* svr)
{
	if (atoi( GetValue(svr,"numplayers") ) && atoi( GetValue(svr, "maxplayers") ) &&
			atoi ( GetValue(svr,"numplayers") ) == atoi( GetValue(svr, "maxplayers") ) )
		return 1;
	return 0;
}
int __stdcall Init(shared_t* info)
{
	types_t* t;

	GetServerAddr = info->GetServerAddr;
	GetServerPort = info->GetServerPort;

	

	HashInit(info->hashes,info->numHashes);
	s = HashSelectInt(HASH_GLOBAL,"Global","Socket");
	clan_id = HashSelectText(HASH_GLOBAL,"Settings","ClanId");
	

	t = info->types[0];
	//t->type = 16;
	t->port = 28000;
	t->protocol = 1;
	strcpy(t->name, "Tribes 2");
	strcpy(t->path, "C:\\Tribes2\\gamedata\\tribes2.exe");
	strcpy(t->qstat_type, "T2S");
	strcpy(t->gamespy_type, "t2prot");
	strcpy(t->mod_var,"gamename");
	
	t->receive = TRIBES2_Receive;
	t->print_info = TRIBES2_PrintInfo;
	t->send = TRIBES2_Send;
	t->hostname = TRIBES2_Hostname;
	t->isfull = TRIBES2_IsFull;

	return 1;
}

