/*
******************************************************************
Server Query; a free games server monitoring tool
Copyright (C) 1998-2000  Nathan O'Sullivan

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
******************************************************************
*/
#include <stdio.h>
#include "types.h"
#include <time.h>
#include "main.h"
#include "list.h"
#include "limits.h"
#include "utils.h"

#include "shared.h"
#include <io.h>
#include "comm.h"
#include "hashdef.h"
#include "default.h"

SOCKET InitSocket(void);
char* strip(char* str);

types_t *server_types[MAX_SERVER_TYPES];
int num_server_types = 0;

types_t *protocol_types[MAX_PROTOCOL_TYPES];
int num_protocol_types = 0;

types_t* GetServerType_qstat(char* type)
{
	int i;
	for (i=0;i<num_server_types;i++)
	{
		if (SQstricmp(server_types[i]->qstat_type,type) == 0)
			return server_types[i];
	}
	return server_types[0];
}

types_t* GetServerType_gamespy(char* type)
{
	int i;
	for (i=0;i<num_server_types;i++)
	{
		if (SQstricmp(server_types[i]->gamespy_type,type) == 0)
			return server_types[i];
	}
	return server_types[0];
}

types_t* GetServerType(int idx)
{
	/*int i;
	for (i=0;i<num_server_types;i++)
	{
		if (server_types[i]->type == type)
			return server_types[i];
	}
	return server_types[0];*/
	if (idx < num_server_types)
		return server_types[idx];
	else
		return server_types[0];
}

void RequestServerStatus(server_t* svr) 
{
	types_t* t;
	if (!svr)
		return;

	if (GetAccTickCount() - svr->last_req < SEND_TIMEOUT)
		return;

	svr->last_req = GetAccTickCount();
	t = svr->gametype;
	t->send(HSIG("Global","Socket"), svr, t->port);
}

extern unsigned int last_req;

void RefreshLANServers(void)
{
	list_t* list;
	list = HashFind(HASH_GLOBAL,"Lists","LANServers", (int) GroupOfList(NULL)->name);
	if (!list)
		return;

	EmptyList(list);

	HashInsertInt(HASH_GLOBAL,"Status","LanServerType",0);
	RequestLANServersStatus();
}


void RequestLANServersStatus(void)
{
	server_t* svr;
	list_t* list;
	char buf[128];
	int i = (int) HashSelect(HASH_GLOBAL,"Status","LanServerType",HASH_INT);
	int limit = i + 1;

	if (!HSIG("Settings","ShowLANList"))
		return;

	WRITE_DEBUG("Finding current type");


	/*if (HashLookupI("Status","IgnoreLAN"))
		return;*/

	/*int i = 0;
	int limit = 16;*/
	
	//limit = num_server_types;


	if (limit > num_server_types)
	{
		WRITE_DEBUG("limit too big, limiting further");
		limit = num_server_types;
	}

	WRITE_DEBUG("finding lan list");
	list = HashFind(HASH_GLOBAL,"Lists","LANServers", (int) GroupOfList(NULL)->name);
	if (!list)
		return;

	
	for (;i<limit;i++)
	{
		WRITE_DEBUG("pinging servers on lan");

		if (server_types[i]->port)
		{
			sprintf(buf,"ping server type '%s' on port %d",
					server_types[i]->qstat_type,server_types[i]->port);
			WRITE_DEBUG(buf);
			server_types[i]->send(server_types[i]->bc_socket, NULL, (short) (server_types[i]->recv_port?server_types[i]->recv_port:server_types[i]->port));		

			HIIG("LastRequest",(HASH_PARAM)server_types[i]->qstat_type, GetAccTickCount() );
		}
		WRITE_DEBUG("lan ping done");

		WRITE_DEBUG("now checking lan servers");
		svr = HashReset(list,"servers",HASH_PTR);
		while (svr) {
			WRITE_DEBUG("checking for current lan servers");
			if (svr->gametype == server_types[i]) 
			{
				WRITE_DEBUG("resetting last request ");
				svr->last_req = GetAccTickCount();
			}				
			svr = HashNext(list,"servers",HASH_PTR);
		}
		WRITE_DEBUG("done checking known lan servers");
	}
	i = limit;

	WRITE_DEBUG("set next request time");
	if (i >= num_server_types) // max server types
	{
		HashInsert(HASH_GLOBAL,"Status","LanServerType",0,HASH_INT);
		HashDelete(HASH_GLOBAL,"Status","LANRefresh");
	} else {
		HashInsert(HASH_GLOBAL,"Status","LanServerType",(HASH_PARAM)i,HASH_INT);
		HashInsert(HASH_GLOBAL,"Status","LANRefresh",(HASH_PARAM)(GetTickCount() + 100),HASH_INT);
	}
	// fix me
	last_req = GetAccTickCount();
	WRITE_DEBUG("returning");
}

void ReadTypeList(void)
{
	FILE* fp;
	char buf[1024];
	char gamespy_type[256];
	char key[1024];
	char value[1024];
	char iconname[256];
	types_t* t = NULL;
	char* idx;
	int i;

	fp = fopen("game\\typelist.txt","rt");
	if (!fp)
		return;

	
	
	while (!feof(fp))
	{
		fgets(buf,1024,fp);
		strip(buf);

		if (buf[0] == '#')
			continue;

		if (buf[0] == '[')
		{
			strcpy(gamespy_type,buf+1);
			gamespy_type[strlen(gamespy_type)-1] = 0; // remove ]
		}

		// explode
		strcpy(key,buf);
		idx = strchr(key,'=');
		if (idx == NULL)
			continue;

		*idx = 0;
		strcpy(value,idx+1);

		// action
		if (stricmp(key,"protocol") == 0)
		{
			// copy the protocol
			for (i=0;i<num_protocol_types;i++)
			{
				if (stricmp(protocol_types[i]->gamespy_type,value) == 0)
					break;
			}
			if (i == num_protocol_types)
				t = NULL;
			else
			{
				t = server_types[num_server_types++];
				memcpy(t,protocol_types[i],sizeof(types_t));
				strcpy(t->gamespy_type,gamespy_type);
				strcpy(t->path,"");

				// set up parameters
				t->parameters = malloc(256);
				strcpy(t->parameters,"");

				// load broadcast socket
				t->bc_socket = (int) InitSocket();


				sprintf(iconname,"game\\%s.bmp",t->gamespy_type);

				(void*)t->iImage = LoadImage(
						NULL, iconname, IMAGE_BITMAP, 0,0, LR_LOADFROMFILE
				);
			}
		}
		else if (t != NULL)
		{
			if (stricmp(key,"name") == 0)
				strcpy(t->name, value);
			else if (stricmp(key,"connect") == 0)
				strcpy(t->connect_template, value);
			else if (stricmp(key,"qstat") == 0)
				strcpy(t->qstat_type,value);
			else if (stricmp(key,"gameport") == 0)
				sscanf(value,"%hd",&t->port);
			else if (stricmp(key,"offset") == 0)
			{
				sscanf(value,"%hd",&t->port_offset);
				t->recv_port = t->port + t->port_offset;
			}
		}
	}
}

void RegisterGameTypes(void)
{
	shared_t info;
	int i, handle, j, adj;
	HMODULE dll;
	struct _finddata_t fd;
	char filename[256];
	char iconname[256];
	types_t* type;

	int (__stdcall *Init)(shared_t*);

	info.GetServerAddr = GetServerAddr;
	info.GetServerPort = GetServerPort;
	info.hashes = G_Hashes;
	info.numHashes = &G_NumHashes;


	num_server_types = 0; 
	for (i=0;i<MAX_SERVER_TYPES;i++)
	{
		server_types[i] = calloc(sizeof(types_t),1);
		server_types[i]->isPacketFrom = DEFAULT_isPacketFrom;
		server_types[i]->mod_cvar = "game";
		HashAlloc(server_types[i],"funcs",8);
	}

	num_protocol_types = 0; 

	SetDefault(server_types[0]);
	num_server_types++;

	
	handle = _findfirst("game/*.dll",&fd);

	if (handle != -1)
	{
		do {
			sprintf(filename,"game/%s",fd.name);

			if ((dll = LoadLibrary(filename)) != NULL)
			{
				HashInsertPtr(NULL,"Status","hInstance",dll);
				info.types = server_types + num_server_types;
				Init = (int (__stdcall *)(shared_t*)) GetProcAddress(dll, "Init");
				if (Init) {
					i = Init(&info);
					server_types[num_server_types]->module = (void*) dll;
					for (j=0;j<i;j++) 
					{
						type = server_types[num_server_types+j];

						// If this is a protocol, move to the protocol list
						if (type->protocol)
						{
							protocol_types[num_protocol_types++] = type;

							for (adj = num_server_types+j; adj < num_server_types+i-1;adj++)
							{
								server_types[adj] = server_types[adj+1];
							}
							server_types[adj] = calloc(sizeof(types_t),1);
							server_types[adj]->isPacketFrom = DEFAULT_isPacketFrom;
							server_types[adj]->mod_cvar = "game";
							HashAlloc(server_types[adj],"funcs",8);

							i--; j--;
							continue;
						}
							
						// load icon if wasnt loaded by DLL
						if (!server_types[num_server_types+j]->iImage)
						{
							sprintf(iconname,"game\\%s.bmp",server_types[num_server_types+j]->gamespy_type);

							(void*)server_types[num_server_types+j]->iImage = LoadImage(
								NULL, iconname, IMAGE_BITMAP, 0,0, LR_LOADFROMFILE
							);

							//sprintf(iconname, "%d", GetLastError());
							//MessageBox(0,iconname,"Load Image",MB_OK);
		
						}

						// set up parameters
						server_types[num_server_types+j]->parameters = malloc(256);
						strcpy(server_types[num_server_types+j]->parameters,"");

						// load broadcast socket
						server_types[num_server_types+j]->bc_socket = (int) InitSocket();
					}
					num_server_types += i;

				}
			}
		} while (! _findnext(handle,&fd) );
		_findclose(handle);
	}

	ReadTypeList();

	for (i=num_server_types;i<MAX_SERVER_TYPES;i++) {
		HashFreeAssoc(server_types[i]);
		free(server_types[i]);
	}
	return;
}
			