/*
******************************************************************
Server Query; a free games server monitoring tool
Copyright (C) 1998-2000  Nathan O'Sullivan

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
******************************************************************
*/
#include <stdio.h>
#include "types.h"
#include "list.h"
#include "main.h"
#include "utils.h"
#include <commctrl.h>

void UI_TreeView_Server(server_t* svr, int insert)
{
	TV_INSERTSTRUCT tv;
	ITEM_TYPE* it;
	char buf[128];
	int idx = 0;
	list_t* list;
	
  for (list = HashReset(svr,"lists",HASH_PTR);list;list = HashNext(svr,"lists",HASH_PTR),idx++)
  {
	HWND hwnd = HashSelectPtr(HASH_GLOBAL,"Global","ServersTree");
	
	if ((it = HashSelectPtr(svr, "Ptrs", "Item_type")) == NULL)
	{
		it = (ITEM_TYPE*) calloc(sizeof(ITEM_TYPE),1);
		it->type = IT_SERVER;
		it->ptr = svr;
		it->owner = list;
		HashInsertPtr(svr,"Ptrs","Item_type",it);
	}

	tv.hParent = list->hTreeItem;
	tv.hInsertAfter = TVI_LAST;

	strcpy(buf,svr->name);
	HashInsertText(svr,"temp","hostname",buf);
	tv.item.pszText = HashSelectText(svr,"temp","hostname");
	tv.item.lParam = (LPARAM)it;
	tv.item.iImage = HashSelectInt(NULL,"Indexes","Blank");
	tv.item.iSelectedImage = HashSelectInt(NULL,"Indexes","Blank");

	if (svr->gametype->iImage) 
		tv.item.iImage = tv.item.iSelectedImage = svr->gametype->iImage;

	tv.item.mask = TVIF_TEXT | TVIF_PARAM | TVIF_IMAGE | TVIF_SELECTEDIMAGE;
	if (insert || HashSelectPtr(svr,"hTreeItems",(HASH_PARAM)idx) == NULL)
		HashInsertPtr(svr,"hTreeItems", (HASH_PARAM)idx, TreeView_InsertItem(hwnd,&tv) );
	else {
		tv.item.hItem = HashSelectPtr(svr,"hTreeItems",(HASH_PARAM)idx);
		TreeView_SetItem(hwnd,&(tv.item));
	}
	
	// FIXME - too hacky
	if (strcmp(GroupOfList(list)->name,"LANServers") == 0)
	{
		TreeView_SortChildren(HashSelect(HASH_GLOBAL,"Global","ServersTree",HASH_PTR),
			HashSelect(HASH_GLOBAL,"HTREEITEM","LANServers",HASH_PTR),0);
	}
  }
}

void UI_ListView_UpdateServer(server_t* svr)
{
	int i,j;
	LV_FINDINFO lvf;
	LV_ITEM lvi;
	list_t* list;
	char buf[128], buf2[128];

	int has_buddy = 0;

	for (j=0;j<HashCount(HASH_GLOBAL,"Buddy");j++)
	{
		strcpy(buf,HashSelectText(HASH_GLOBAL,"Buddy",(HASH_PARAM)j));
		SQstrtoupper(buf);

		for (i=0;i<HashSelectInt(svr,"stats","numplayers");i++)
		{
			strcpy(buf2,HashSelectText(svr,"name",(HASH_PARAM)i) );
			SQstrtoupper(buf2);

			if (strstr(buf2,buf) != NULL)
			{
				has_buddy = 1;
				break;
			}
		}
		if (has_buddy)
			break;
	}
	

  for (list=HashReset(svr,"lists",HASH_PTR);list;list=HashNext(svr,"lists",HASH_PTR))
  {
	if (list != HashSelectPtr(HASH_GLOBAL,"Status","CurrentList"))
		continue;

	lvf.flags = LVFI_PARAM;
	lvf.lParam = (long)svr;
	i = ListView_FindItem(HSPG("Global","ServersLV"),-1,&lvf);

	/* No idea why, but obviously this triggers an item change */
	HIIG("Status","IgnoreItemChange",1);
	/*ListView_SetItemText(HSPG("Global","ServersLV"),
		i,0,svr->name);*/

	lvi.mask = LVIF_TEXT | LVIF_PARAM | LVIF_IMAGE | LVIF_STATE;
	if (svr->gametype->iImage) 
		lvi.iImage = svr->gametype->iImage;
	else 
		lvi.iImage = 0;
			
	lvi.iItem = i;
	lvi.iSubItem = 0;
	lvi.state = lvi.stateMask = 0;
	lvi.state = 7 << 12;
	lvi.stateMask = INDEXTOSTATEIMAGEMASK(has_buddy<<1);

	lvi.pszText = svr->name;
	lvi.lParam = (long)svr; //LPSTR_TEXTCALLBACK; //(long)svr;
	ListView_SetItem(HashSelectPtr(HASH_GLOBAL,"Global","ServersLV"),&lvi);
	
	HIIG("Status","IgnoreItemChange",0);

	/* Redraw the sub items */
	ListView_RedrawItems(HashSelectPtr(HASH_GLOBAL,"Global","ServersLV"),i,i);
  }
}

void UI_ListView_AddServer(server_t* svr,list_t* list)
{
	LV_ITEM lvi;
	int i;
    //list_t* list;

  /*for (list = HashReset(svr,"lists",HASH_PTR);list;list = HashNext(svr,"lists",HASH_PTR))
  { */
	if (list != CURRENT_LIST)
	//	continue;
		return;

	i = ListView_GetItemCount(HSPG("Global","ServersLV"));

	lvi.mask = LVIF_TEXT | LVIF_PARAM | LVIF_IMAGE;
	if (svr->gametype->iImage) 
		lvi.iImage = svr->gametype->iImage;
	else 
		lvi.iImage = 0;
			
	lvi.iItem = i;
	lvi.iSubItem = 0;
	lvi.state = lvi.stateMask = 0;
	lvi.pszText = svr->name;
	lvi.lParam = (long)svr; //LPSTR_TEXTCALLBACK; //(long)svr;
	ListView_InsertItem(HashSelectPtr(HASH_GLOBAL,"Global","ServersLV"),&lvi);
  //}
}

void UI_AddServer(server_t* svr, list_t* list)
{
	if (HSIG("Settings","DualPane") == 0)
		UI_TreeView_Server(svr,1);
	else UI_ListView_AddServer(svr,list);
}

void UI_UpdateServer(server_t* svr)
{
if (HSIG("Settings","DualPane") == 0)
		UI_TreeView_Server(svr,0);
	else UI_ListView_UpdateServer(svr);
}

void UI_TreeView_DeleteServer(server_t* server, list_t* list)
{
	void* hTreeItem;
	for (hTreeItem = HashReset(server,"hTreeItems",HASH_PTR);hTreeItem;hTreeItem=HashNext(server,"hTreeItems",HASH_PTR))
	{
		if (hTreeItem != TVI_ROOT && hTreeItem) {
			TreeView_DeleteItem(HashSelect(HASH_GLOBAL,"Global","ServersTree",HASH_PTR),
				hTreeItem);
		}
	}
	HashDeleteAll(server,"hTreeItems");
}

void UI_ListView_DeleteServer(server_t* server, list_t* list)
{
	LV_FINDINFO lvf;
	int i;
	//list_t* list;

  /*for (list=HashReset(server,"lists",HASH_PTR);list;list=HashNext(server,"lists",HASH_PTR))
  {	*/  
	if (list == CURRENT_LIST)
	{
		lvf.flags = LVFI_PARAM;
		lvf.lParam = (long)server;
		i = ListView_FindItem(HSPG("Global","ServersLV"),-1,&lvf);

		if (i != -1)
			ListView_DeleteItem(
				HashSelectPtr(HASH_GLOBAL,"Global","ServersLV"),
				i
			);
	}
  //}
}

void UI_DeleteServer(server_t* svr, list_t* list)
{
	if (HSIG("Settings","DualPane") == 0)
		UI_TreeView_DeleteServer(svr,list);
	else UI_ListView_DeleteServer(svr,list);
}

int CALLBACK UI_SortServersLVCompare(LPARAM p1, LPARAM p2, LPARAM misc)
{
	server_t* s1, *s2;
	int i = HSIG("Settings","ServersLV-SortType");
	int a, b;
	char *t1, *t2;

	s1 = (server_t*)p1;
	s2 = (server_t*)p2;

	// unsorted
	if (i == 0)
	{
		a = (int)HashExistsValue((void*)misc,"servers",(void*)p1,HASH_PTR);
		b = (int)HashExistsValue((void*)misc,"servers",(void*)p2,HASH_PTR);
		if (a < b)
			return -1;
		if (a > b)
			return 1;
		return 0;
	}
	// sort by name
	else if (i == 1)
		return strcmp(s1->name,s2->name);
	// sort by players, highest at top
	else if (i == 2)
	{
		a = HSI(s1,"stats","numplayers");
		b = HSI(s2,"stats","numplayers");
		if (a < b)
			return 1;
		if (a > b)
			return -1;
		a = HSI(s1,"stats","maxplayers");
		b = HSI(s2,"stats","maxplayers");
		if (a < b)
			return -1;
		if (a > b)
			return 1;
		return strcmp(s1->name,s2->name);
	}
	// sort by ping
	else if (i == 3)
	{
		a = s1->ping;
		b = s2->ping;

		if (a == 0)
			a = 10000;
		if (b == 0)
			b = 10000;

		if (a<b)
			return -1;
		else if (a>b)
			return 1;
		else return strcmp(s1->name,s2->name);
	}
	// sort by avg ping
	else if (i == 4)
	{
		a = s1->avg_ping;
		b = s2->avg_ping;

		if (a == 0)
			a = 10000;
		if (b == 0)
			b = 10000;

		if (a<b)
			return -1;
		else if (a>b)
			return 1;
		else return strcmp(s1->name,s2->name);
	}
	// sort by map
	else if (i == 5)
	{
		t1 = HST(s1,"stats","map");
		t2 = HST(s2,"stats","map");
		
		if (strcmp(t1,"") == 0)
			t1 = "\x7f\x7f\x7f\x7f";
		if (strcmp(t2,"") == 0)
			t2 = "\x7f\x7f\x7f\x7f";

		return strcmp(t1,t2);
	}
	else if (i == 6)
	{
		a = strcmp(s1->gametype->name,
			s2->gametype->name);

		if (!a)
			return strcmp(s1->name, s2->name);
		else 
			return a;
	}
	return 0;		
}

void UI_SortList(list_t* list)
{
	if (HSIG("Settings","DualPane") == 1 && list == CURRENT_LIST)
		ListView_SortItems(HSPG("Global","ServersLV"),UI_SortServersLVCompare,list);
}




void SaveWindowPosition(HWND hwnd) {
	WINDOWPLACEMENT wp;
	int i = SW_SHOWMAXIMIZED, j, k;
	char buf[128];
	char* windows[] = { "PlayerInfoLV","ServerInfoLV","ServersLV", NULL };

		
	wp.length = sizeof(WINDOWPLACEMENT);
	GetWindowPlacement(hwnd,&wp);
	HashInsertInt(HASH_GLOBAL,"Settings","WindowLeft",wp.rcNormalPosition.left);
	HashInsertInt(HASH_GLOBAL,"Settings","WindowTop",wp.rcNormalPosition.top);
	HashInsertInt(HASH_GLOBAL,"Settings","WindowWidth",wp.rcNormalPosition.right-wp.rcNormalPosition.left + 1);
	HashInsertInt(HASH_GLOBAL,"Settings","WindowHeight",wp.rcNormalPosition.bottom-wp.rcNormalPosition.top + 1);
	HashInsertInt(HASH_GLOBAL,"Settings","WindowMax",(wp.showCmd & SW_SHOWMAXIMIZED) == SW_SHOWMAXIMIZED);

	// save listview column widths
	for (j=0;windows[j];j++)
	{
		hwnd = HSPG("Global",windows[j]);
		if (hwnd)
		{
			for (i=0;i<10;i++)
			{
				k = ListView_GetColumnWidth(hwnd,i);
				if (!k)
					break;
				sprintf(buf,"%s-Col%d",windows[j],i);
				HIIG("Settings",buf,k);
			}
		}
	}

}

void LoadWindowPosition(HWND hwnd,int def) {
	int x, y, w, h;
	HWND screen;
	RECT rect;
	
	// for listviews
	char buf[128];
	char* windows[] = { "PlayerInfoLV","ServerInfoLV","ServersLV", NULL };
	int i,j,k;
	//

	screen = GetDesktopWindow();
	GetWindowRect(screen,&rect);

	if (HashSelectInt(HASH_GLOBAL,"Settings","WindowWidth")) {
		x = HashSelectInt(HASH_GLOBAL,"Settings","WindowLeft");
		y = HashSelectInt(HASH_GLOBAL,"Settings","WindowTop");
		w = HashSelectInt(HASH_GLOBAL,"Settings","WindowWidth");
		h = HashSelectInt(HASH_GLOBAL,"Settings","WindowHeight");

		if (x < 0 || x >= rect.right) x = 0;
		if (y < 0 || y >= rect.bottom) y = 0;
		if (w < 50 || w > 3000)
			w = 500;
		if (h < 50 || h > 2000)
			h = 380;

		MoveWindow(PROG_HWND, x, y, w, h, 0);
	}

	if (HashSelectInt(HASH_GLOBAL,"Settings","WindowMax"))
		ShowWindow(PROG_HWND,SW_SHOWMAXIMIZED);
	else ShowWindow(PROG_HWND,def);

	// load listview column widths
	for (j=0;windows[j];j++)
	{
		hwnd = HSPG("Global",windows[j]);
		if (hwnd)
		{
			for (i=0;i<10;i++)
			{
				sprintf(buf,"%s-Col%d",windows[j],i);
				k = HSIG("Settings",buf);
				if (k <= 0 || k > 10000)
					break;
				ListView_SetColumnWidth(hwnd,i,k);
			}
		}
	}
}