/*
******************************************************************
Server Query; a free games server monitoring tool
Copyright (C) 1998-2002  Nathan O'Sullivan

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
******************************************************************
*/
#include <stdio.h>
#include "resource.h"
#include "unreal\resource.h"
#include "types.h"
#include "shared.h"
#include "players.h"
#include "servinfo.h"

#define TYPE_UNR	7
#define TYPE_UT		10
#define TYPE_RUNE	11
#define TYPE_GORE	15
#define TYPE_MOHAADEMO  19
#define TYPE_MOHAA  20
#define TYPE_MOHAA10  21
#define TYPE_SSAM	22
#define TYPE_SSAM2	23
#define TYPE_SSAM2DEMO	24
#define TYPE_RENEGADEDEMO 25
#define TYPE_GLOBALOPSDEMO 26
#define TYPE_RENEGADE 27
#define TYPE_TACOPS	38

#define DEFAULT_PORT 7777

#define SVR_NUMPLAYERS (HSI(svr,"info","num_players"))
#define SVR_FRAGS(x) (HSI(svr,"frags",(HASH_PARAM)x))
#define SVR_NAME(x) (HST(svr,"name",(HASH_PARAM)x))
#define SVR_PING(x) (HSI(svr,"ping",(HASH_PARAM)x))
#define SVR_CHANGE(x) (HSI(svr,"change",(HASH_PARAM)x))

SOCKET s;
char* clan_id;
unsigned long (*GetServerAddr)(char* addr);
short int (*GetServerPort)(char* addr, short otherwise);
types_t*		(*MAIN_GetServerType)		(int type);
types_t*		(*MAIN_GetServerType_qstat)		(char* type);

//FIXME: this is largely the code from quakex.c, only a few modifications have been made

void UNREAL_DecodeString(server_t* svr, char* str, int len)
{
	int which, pos;
	int i = 0, cur_player;
	//FIXME: this is lazy, there should be bounds checking
	char key[1024], value[1024];
	int itemlen = 4096;

	char* ptr = str;
	
	if (svr->WAITING_FIRST_PACKET) {
		HashDeleteAll(svr,"items");
		HII(svr,"info","num_players",0);
	}
	svr->WAITING_FIRST_PACKET = 0;

	pos = 0; which = -1; 
	
	do {
		if (*ptr == '\\' || *(ptr+1) == 0)
		{
			if (*(ptr+1) == 0)
			{
				if (which == 0) key[pos] = *ptr;
				else value[pos] = *ptr;
				pos++;
			}
			
			if (which == 0) {
				key[pos] = 0;
				which = 1;
			}
			else if (which == 1) {
				value[pos] = 0;
				which = 0;

				if (strcmp(key,"servername") == 0)
					HashInsertText(svr,"items","hostname",value);
				else if (!memcmp(key,"player_",7)) {
					

					// this is kind of poor form - we take a punt
					// that cvars are returned grouped by player id
					cur_player = SVR_NUMPLAYERS;
					HII(svr,"info","num_players",SVR_NUMPLAYERS + 1);
					

					HIT(svr,"name",(HASH_PARAM)cur_player,value);
				}
				else if (!memcmp(key,"frags_",6)) 
					HIT(svr,"frags",(HASH_PARAM)cur_player,value);
				else if (!memcmp(key,"score_",6)) 
					HIT(svr,"frags",(HASH_PARAM)cur_player,value);
				else if (!memcmp(key,"deaths_",6)) 
					HIT(svr,"deaths",(HASH_PARAM)cur_player,value);
					//svr->players[atoi(key + 6)].frags = atoi(value);
				else if (!memcmp(key,"ping_",5)) 
					HIT(svr,"ping",(HASH_PARAM)cur_player,value);
					//svr->players[atoi(key + 5)].ping = atoi(value);
				else if (!memcmp(key,"team_",5) || !memcmp(key,"skin_",5) ||
					!memcmp(key,"mesh_",5))
					; // dont add to hash, dunno what to do with them
				else 
					HashInsertText(svr,"items",key,value);
	
			}
			else which = 0;
			pos = 0;
		}/*
		else if (*ptr == '\x0a') 
		{
			value[pos] = 0;
			HashInsertText(svr,"items",key,value); //fixme : never called
		}*/
		else if (pos < itemlen) {
			if (which == 0) key[pos] = *ptr;
			else value[pos] = *ptr;
			pos++;
		}
		ptr++;
	} while (*ptr != 0 && ptr-str < len);
	
	CleanRecords(svr);

	return;
}

void UNREAL_Send(SOCKET s, server_t* svr, short port) 
{
	SOCKADDR_IN src_addr;
	types_t* svrtype;
	int diff;

	if (svr)
	{
		//svrtype = MAIN_GetServerType(svr->type);
		svrtype = svr->gametype;
		diff = svrtype->recv_port - svrtype->port;
	}
	else diff = 0; // have to assume we've already been given the 'right' port
	
	src_addr.sin_family = AF_INET;
	if (svr) {
		src_addr.sin_addr.s_addr = GetServerAddr(svr->ipaddr);
		src_addr.sin_port = htons((short)(GetServerPort(svr->ipaddr, port) + diff ) );
	}

	else {
		src_addr.sin_addr.s_addr = INADDR_BROADCAST;
		src_addr.sin_port = htons((short)(port) );
	}

	// UNREAL monitors game port + 1 for status packets
	
	sendto(s, "\\status\\", 8, 0, (SOCKADDR*)&src_addr, sizeof(SOCKADDR_IN));
	return;
}

void UNREAL_PrintInfo(server_t* svr, HWND mainwnd, int ignore_changes)
{
//	player_t temp_p;
	HWND hwnd;
	int i, j, tmp;
	char buf[128];
	int sin = 0;
	int order[128];
	types_t* st;

	hwnd = mainwnd;
	ServinfoInit(svr);

	strcpy(buf, GetValue(svr, "hostname") );
	if (svr->timeout >= 5) {
		strcpy(buf, "UNREACHABLE");
		HashDeleteAll(svr,"items"); // wipe all current settings so we get blanks
		HII(svr,"info","num_players",0);
	} else if (HashCount(svr,"items") == 0) {
		HII(svr,"info","num_players",0);
		strcpy(buf,"CONNECTING...");
	}
	//SetDlgItemText(hwnd, IDT_HOSTNAME, buf );
	ServinfoAdd("Hostname",buf);
	ServinfoAdd("Address", svr->ipaddr);

		//FIXME
	st = svr->gametype;
	/*
	switch(svr->type)
	{
		case TYPE_UT:
			strcpy(buf, "UT");
			break;
		case TYPE_MOHAA:
		case TYPE_MOHAA10:
			strcpy(buf, "MOH:AA");
			break;
		case TYPE_MOHAADEMO:
			strcpy(buf, "MOH:AA Demo");
			break;
		default: if (st) strcpy(buf,st->name); else strcpy(buf,"GameSpy Protocol");
	}*/
	if (st) strcpy(buf,st->name); else strcpy(buf,"GameSpy Protocol");

	if (strcmp(GetValue(svr,"gamever"),""))
				sprintf(buf+strlen(buf)," v%s",GetValue(svr,"gamever"));

	//SetDlgItemText(hwnd, IDT_VERSION, buf);
	ServinfoAdd("Game",buf);

	strcpy(buf, GetValue(svr, "mapname") );
	if (!strcmp( GetValue(svr, "mapname"), "") || GetValue(svr, "mapname") == NULL)
	{
		strcpy(buf, GetValue(svr, "map") );
	}
	//SetDlgItemText(hwnd, IDT_MAPNAME, buf );
	ServinfoAdd("Map", buf);

	if (strcmp(GetValue(svr, "maxplayers"), ""))
	{
		sprintf(buf, "%s / %s", GetValue(svr,"numplayers"), GetValue(svr, "maxplayers") );
	}
	else strcpy(buf, "");
	//SetDlgItemText(hwnd, IDT_PLAYERS, buf);
	ServinfoAdd("Players",buf);

	ServinfoAdd("Ping","");
	
  if (HSIG("Settings","ExtraServInfo")) 
  {
	strcpy(buf, GetValue(svr, "fraglimit") );
	//SetDlgItemText(hwnd, IDT_FRAGLIMIT, buf );
	ServinfoAdd("Fraglimit",buf);

	strcpy(buf, GetValue(svr, "timelimit") );
	//SetDlgItemText(hwnd, IDT_TIMELIMIT, buf );
	ServinfoAdd("Timelimit",buf);

	//SetDlgItemText(hwnd, IDC_CAPTURELIMIT, "Capturelimit:");
	strcpy(buf, GetValue(svr, "capturelimit") );
	//SetDlgItemText(hwnd, IDT_CAPTURELIMIT, buf );
	ServinfoAdd("Capturelimit",buf);

/*	SetDlgItemText(hwnd, IDC_DMFLAGS, "DM Flags:" );
	strcpy(buf, GetValue(svr, "dmflags") );
	SetDlgItemText(hwnd, IDT_DMFLAGS, buf);
	ServinfoAdd("DM Flags",buf);*/

	//SetDlgItemText(hwnd,IDT_MISC,"");
  }
    PlayerinfoInit(svr);
	//temp = GetDlgItem(hwnd, IDL_PLAYERS);

	//SendMessage(temp, LB_RESETCONTENT, 0, 0);
	
	for (i=0;i<SVR_NUMPLAYERS;i++)
		order[i] = i;

	i = 0;
	while (i < SVR_NUMPLAYERS - 1) 
	{
		if (SVR_FRAGS(order[i]) < SVR_FRAGS(order[i+1]))
		{
			j = order[i];
			order[i] = order[i+1];
			order[i+1] = j;
			//temp_p = svr->players[i];
			//svr->players[i] = svr->players[i+1];
			//svr->players[i+1] = temp_p;
			i = 0;
			continue;
		}
		i++;
	}
	
	//SendMessage(temp, LB_ADDSTRING, 0, (LPARAM)"\tFrags\tPing\tName");
		
	for (i=0;i<SVR_NUMPLAYERS;i++)
	{
		tmp = SVR_NUMPLAYERS;

		strcpy(buf,"");

		if (strcmp(clan_id, ""))
		{
			for (j=0;j<(int)(strlen(SVR_NAME(order[i])) - strlen(clan_id));j++)
				if (!memcmp(SVR_NAME(order[i])+j, clan_id, strlen(clan_id)))
					strcat(buf,"*");
		}
		
		strcat(buf, "\t");

		if (!ignore_changes && SVR_CHANGE(order[i]))
			sprintf(buf + strlen(buf), "%d (%c%d)\t%d\t%s", 
				SVR_FRAGS(order[i]),
				SVR_CHANGE(order[i]) > 0 ? '+' : ' ',
				SVR_CHANGE(order[i]),
				SVR_NAME(order[i])
			);
			
		else 
			sprintf(buf + strlen(buf), "%d\t%d\t%s", 
				SVR_FRAGS(order[i]),
				SVR_PING(order[i]),
				SVR_NAME(order[i])
			);
		
		//SendMessage(temp, LB_ADDSTRING, 0, (LPARAM)buf);
		
		PlayerinfoAdd(SVR_NAME(order[i]),SVR_FRAGS(order[i]),
				SVR_CHANGE(order[i]),SVR_PING(order[i]));
		
	}
	return;
}

char* UNREAL_Hostname(server_t* svr)
{
	return GetValue(svr, "hostname");
}

int UNREAL_Receive(server_t* svr, char* buf, int len)
{
	int i,j;

	buf[len] = 0;
	UNREAL_DecodeString(svr, buf, len);

	i = SVR_NUMPLAYERS;
	j = atoi(GetValue(svr,"numplayers"));

	
	if ((strcmp(svr->gametype->qstat_type,"GRS")==0 || GetValue(svr,"hostname")[0]) ) // && SVR_NUMPLAYERS == atoi(GetValue(svr,"numplayers")))
	{
		/*if (svr->type == TYPE_UNR)
		{*/
			if (strcmp(GetValue(svr,"gamename"),"serioussam") == 0)
				svr->gametype = MAIN_GetServerType_qstat("SMS");
			if (strcmp(GetValue(svr,"gamename"),"serioussamse") == 0)
				svr->gametype = MAIN_GetServerType_qstat("S2S");
			if (strcmp(GetValue(svr,"gamename"),"serioussamsed") == 0)
				svr->gametype = MAIN_GetServerType_qstat("S2d");
			if (strcmp(GetValue(svr,"gamename"),"mohaad") == 0)
				svr->gametype = MAIN_GetServerType_qstat("MHD");
			if (strcmp(GetValue(svr,"gamename"),"mohaa") == 0)
			{
				//if (strcmp(GetValue(svr,"gamever"),"1.00") == 0)
				//	svr->type = TYPE_MOHAA10;
				//else
					svr->gametype = MAIN_GetServerType_qstat("MHS");
			}
			if (strcmp(GetValue(svr,"gamename"),"rune") == 0)
				svr->gametype = MAIN_GetServerType_qstat("RNS");
			else if (strcmp(GetValue(svr,"gamename"),"ut") == 0)
			{
				if (strcmp(GetValue(svr,"gametype"),"s_SWATGame") == 0)
					svr->gametype = MAIN_GetServerType_qstat("TCS");
				else
					svr->gametype = MAIN_GetServerType_qstat("UTS");
			}
			
		//	else svr->type = TYPE_UNR;
		//}

		// Update available information
		HashInsertText(svr,"stats","map",
			HashSelectText(svr,"items","mapname"));
		HashInsertText(svr,"stats","numplayers",
			GetValue(svr,"numplayers") );
		HashInsertText(svr,"stats","maxplayers",
		GetValue(svr,"maxplayers") );

		return 1; // update displayed info
	}
	return 0;
}

int UNREAL_isPacketFrom(server_t* svr, SOCKADDR* addr)
{
	SOCKADDR_IN* dest_addr = (SOCKADDR_IN*)addr;
	int port = DEFAULT_PORT, diff, tmp;
	types_t* svrtype;
	svrtype = svr->gametype;
	diff = svrtype->recv_port - svrtype->port;
	port = svrtype->port;

	tmp = ntohs(dest_addr->sin_port);
	
	/*if (svr->type == TYPE_GORE)
		port = 10051;*/

	if (dest_addr->sin_addr.s_addr == GetServerAddr(svr->ipaddr)
		&& dest_addr->sin_port == htons((short) (GetServerPort(svr->ipaddr, port ) + diff ) ))
		return 1;

	return 0;
}

int UNREAL_IsFull(server_t* svr)
{
	if (atoi( GetValue(svr,"numplayers") ) && atoi( GetValue(svr, "maxplayers") ) &&
			atoi ( GetValue(svr,"numplayers") ) == atoi( GetValue(svr, "maxplayers") ) )
		return 1;
	return 0;
}

char* UNREAL_Cmdline(server_t* svr)
{
	static char string[256];
	strcpy(string, "");

	// If there's a password, set it
	//if (strcmp(HashSelect(HASH_GLOBAL,"Status","Password",HASH_STRING),"") != 0)
	//	sprintf(string+strlen(string)," +password \"%s\"",HashSelect(HASH_GLOBAL,"Status","Password",HASH_STRING));

	sprintf (string+strlen(string),"%s", svr->ipaddr);

	if (strcmp(HSTG("Status","Password"),"") != 0)
		sprintf(string+strlen(string),"?password=%s",HSTG("Status","Password"));

	return string;
}

char* Q3STYLE_Cmdline(server_t* svr)
{
	static char string[256];
	strcpy(string, "");

	// If there's a password, set it
	//if (strcmp(HashSelect(HASH_GLOBAL,"Status","Password",HASH_STRING),"") != 0)
	//	sprintf(string+strlen(string)," +password \"%s\"",HashSelect(HASH_GLOBAL,"Status","Password",HASH_STRING));

	sprintf (string+strlen(string),"+connect %s", svr->ipaddr);
	return string;
}

// BF1942, ie we are dumb programmers
char* BF1942_Cmdline(server_t* svr)
{
	static char string[256];
	strcpy(string, "");

	// If there's a password, set it
	//if (strcmp(HashSelect(HASH_GLOBAL,"Status","Password",HASH_STRING),"") != 0)
	//	sprintf(string+strlen(string)," +password \"%s\"",HashSelect(HASH_GLOBAL,"Status","Password",HASH_STRING));

	sprintf (string+strlen(string),"+restart 1 +joinServer %s +isInternet %d", svr->ipaddr,svr->lan?0:1);
	return string;
}

char* GORE_Cmdline(server_t* svr)
{
	static char string[256];
	strcpy(string, "");

	// If there's a password, set it
	//if (strcmp(HashSelect(HASH_GLOBAL,"Status","Password",HASH_STRING),"") != 0)
	//	sprintf(string+strlen(string)," +password \"%s\"",HashSelect(HASH_GLOBAL,"Status","Password",HASH_STRING));

	sprintf (string+strlen(string),"--connect %s", svr->ipaddr);
	return string;
}

char* VIETCONG_Cmdline(server_t* svr)
{
	static char string[256];
	strcpy(string, "-gn -pn");

	if (strchr(svr->ipaddr,':') == NULL)
		sprintf (string+strlen(string)," -ip %s -hostport 5425", svr->ipaddr);
	else
	{
		char ip[64]; int port;
		sscanf(svr->ipaddr,"%[^:]:%d",ip,&port);
		sprintf(string+strlen(string)," -ip %s -hostport %d",ip,port);
	}

	// If there's a password, set it
	if (strcmp(HashSelect(HASH_GLOBAL,"Status","Password",HASH_STRING),"") != 0)
		sprintf(string+strlen(string)," -password \"%s\"",HashSelect(HASH_GLOBAL,"Status","Password",HASH_STRING));

	return string;
}

char* IGI2_Cmdline(server_t* svr)
{
	static char string[256];
	strcpy(string, "");

	if (strchr(svr->ipaddr,':') == NULL)
		sprintf (string+strlen(string)," ip%s port26001", svr->ipaddr);
	else
	{
		char ip[64]; int port;
		sscanf(svr->ipaddr,"%[^:]:%d",ip,&port);
		sprintf(string+strlen(string)," ip%s port%d",ip,port);
	}

	// If there's a password, set it
	if (strcmp(HashSelect(HASH_GLOBAL,"Status","Password",HASH_STRING),"") != 0)
		sprintf(string+strlen(string)," \"password%s\"",HashSelect(HASH_GLOBAL,"Status","Password",HASH_STRING));

	return string;
}


int __stdcall Init(shared_t* info)
{
	types_t* t, *v;
	int type = 0;

	GetServerAddr = info->GetServerAddr;
	GetServerPort = info->GetServerPort;

	HashInit(info->hashes,info->numHashes);
	s = HashSelectInt(HASH_GLOBAL,"Global","Socket");
	clan_id = HashSelectText(HASH_GLOBAL,"Settings","ClanId");
	MAIN_GetServerType = HashSelectPtr(HASH_GLOBAL,"Global","GetServerType");
	MAIN_GetServerType_qstat = HashSelectPtr(HASH_GLOBAL,"Global","GetServerType_qstat");
	
	t = info->types[type++];
	//t->type = TYPE_UNR;
	t->port = 7777;
	t->recv_port = t->port + 1;
	strcpy(t->name, "Unreal");
	strcpy(t->path, "C:\\unreal\\system\\unr.exe");
	strcpy(t->qstat_type, "UNS");
	strcpy(t->gamespy_type, "unreal");
	
	t->receive = UNREAL_Receive;
	t->print_info = UNREAL_PrintInfo;
	t->send = UNREAL_Send;
	t->isPacketFrom = UNREAL_isPacketFrom;
	t->hostname = UNREAL_Hostname;
	//t->cmdline = UNREAL_Cmdline;
	t->disallowPath = 0;
	t->isfull = UNREAL_IsFull;

	(void*)t->iImage = LoadBitmap(
		HashSelectPtr(HASH_GLOBAL,"Status","hInstance"),
		MAKEINTRESOURCE(IDB_GAMEICON)
	);

	v = info->types[type++];
//	v->type = TYPE_RUNE;
	v->port = t->port;
	v->recv_port = t->recv_port;
	strcpy(v->name, "Rune");
	strcpy(v->path, "C:\\Rune\\System\\Rune.exe");
	strcpy(v->qstat_type, "RNS");
	strcpy(v->gamespy_type, "rune");
	
	v->receive = t->receive;
	v->print_info = t->print_info;
	v->send = t->send;
	v->isPacketFrom = t->isPacketFrom;
	v->hostname = t->hostname;
	//v->cmdline = t->cmdline;
	v->disallowPath = t->disallowPath;
	v->isfull = t->isfull;
	
	(void*)v->iImage = LoadBitmap(
		HashSelectPtr(HASH_GLOBAL,"Status","hInstance"),
		MAKEINTRESOURCE(IDB_GAMEICON3)
	);


	v = info->types[type++];
//	v->type = TYPE_GORE;
	v->port = 10051;
	v->recv_port = v->port+1;
	strcpy(v->name, "Gore");
	strcpy(v->path, "C:\\Program files\\Gore\\Gore.exe");
	strcpy(v->qstat_type, "GRS");
	strcpy(v->gamespy_type, "gore");
	
	v->receive = t->receive;
	v->print_info = t->print_info;
	v->send = t->send;
	v->isPacketFrom = t->isPacketFrom;
	v->hostname = t->hostname;
	//v->cmdline = GORE_Cmdline;
	v->disallowPath = t->disallowPath;
	v->isfull = t->isfull;

	(void*)v->iImage = LoadBitmap(
		HashSelectPtr(HASH_GLOBAL,"Status","hInstance"),
		MAKEINTRESOURCE(IDB_GAMEICON4)
	);

	
	v = info->types[type++];
//	v->type = TYPE_SSAM;
	v->port = 25600;
	v->recv_port = v->port+1;
	strcpy(v->name, "Serious Sam");
	strcpy(v->path, "C:\\Program files\\Serious Sam\\bin\\SeriousSam.exe");
	strcpy(v->qstat_type, "SMS");
	strcpy(v->gamespy_type, "serioussam");
	
	v->receive = t->receive;
	v->print_info = t->print_info;
	v->send = t->send;
	v->isPacketFrom = t->isPacketFrom;
	v->hostname = t->hostname;
	//v->cmdline = Q3STYLE_Cmdline;
	v->disallowPath = t->disallowPath;
	v->isfull = t->isfull;

	(void*)v->iImage = LoadBitmap(
		HashSelectPtr(HASH_GLOBAL,"Status","hInstance"),
		MAKEINTRESOURCE(IDB_GAMEICON6)
	);


	v = info->types[type++];
//	v->type = TYPE_SSAM2;
	v->port = 25600;
	v->recv_port = v->port+1;
	strcpy(v->name, "Serious Sam 2");
	strcpy(v->path, "C:\\Program files\\Serious Sam 2\\bin\\SeriousSam2.exe");
	strcpy(v->qstat_type, "S2S");
	strcpy(v->gamespy_type, "serioussam2");
	
	v->receive = t->receive;
	v->print_info = t->print_info;
	v->send = t->send;
	v->isPacketFrom = t->isPacketFrom;
	v->hostname = t->hostname;
	//v->cmdline = Q3STYLE_Cmdline;
	v->disallowPath = t->disallowPath;
	v->isfull = t->isfull;
	

	(void*)v->iImage = LoadBitmap(
		HashSelectPtr(HASH_GLOBAL,"Status","hInstance"),
		MAKEINTRESOURCE(IDB_GAMEICON6)
	);

	// protocol
	v = info->types[type++];
//	v->type = 0;
	v->protocol = 1;
	v->port = 7777;
	v->recv_port = 7778;
	v->port_offset = 1;
	strcpy(v->gamespy_type, "gsprot");
	
	v->receive = t->receive;
	v->print_info = t->print_info;
	v->send = t->send;
	v->isPacketFrom = t->isPacketFrom;
	v->hostname = t->hostname;
	v->disallowPath = t->disallowPath;
	v->isfull = t->isfull;

	

	return type;
}

